from datetime import datetime
from io import BytesIO
from pathlib import Path
import tempfile
from typing import Dict, List, Optional, Union

from benchling_api_client.v2.stable.api.data_frames import create_data_frame, get_data_frame, patch_data_frame
from benchling_api_client.v2.stable.models.data_frame import DataFrame
from benchling_api_client.v2.stable.models.data_frame_create import DataFrameCreate
from benchling_api_client.v2.stable.models.data_frame_create_manifest_manifest_item import (
    DataFrameCreateManifestManifestItem,
)
from benchling_api_client.v2.stable.models.data_frame_update import DataFrameUpdate
from benchling_api_client.v2.stable.models.data_frame_update_upload_status import DataFrameUpdateUploadStatus
from benchling_api_client.v2.stable.models.file_status_upload_status import FileStatusUploadStatus
from benchling_api_client.v2.types import Response
import httpx

from benchling_sdk.errors import DataFrameInProgressError, InvalidDataFrameError, raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import GetDataFrameRowDataFormat
from benchling_sdk.services.v2.base_service import BaseService

_DEFAULT_HTTP_TIMEOUT_UPLOAD_DATA_FRAME: float = 60.0


class DataFrameService(BaseService):
    """
    Data Frames.

    Data Frames are Benchling objects that represent tabular data with typed columns and rows of data.

    See https://benchling.com/api/v2/reference#/Data%20Frames
    """

    @api_method
    def get_by_id(
        self,
        data_frame_id: str,
        row_data_format: Optional[GetDataFrameRowDataFormat] = None,
        returning: Optional[str] = None,
    ) -> DataFrame:
        """
        Get a data frame and URLs to download its data.

        See https://benchling.com/api/v2/reference#/Data%20Frames/getDataFrame
        """
        response = get_data_frame.sync_detailed(
            client=self.client,
            data_frame_id=data_frame_id,
            returning=none_as_unset(returning),
            row_data_format=none_as_unset(row_data_format),
        )
        return model_from_detailed(response)

    @api_method
    def create(self, data_frame: DataFrameCreate) -> DataFrame:
        """
        Create a data frame.

        See https://benchling.com/api/v2/reference#/Data%20Frames/createDataFrame
        """
        response = create_data_frame.sync_detailed(client=self.client, json_body=data_frame)
        return model_from_detailed(response)

    @api_method
    def update(self, data_frame_id: str, data_frame: DataFrameUpdate) -> TaskHelper[DataFrame]:
        """
        Update a data frame.

        See https://benchling.com/api/v2/reference#/Data%20Frames/patchDataFrame
        """
        response = patch_data_frame.sync_detailed(
            client=self.client, data_frame_id=data_frame_id, json_body=data_frame
        )
        return self._task_helper_from_response(response, DataFrame)

    def upload_bytes(
        self,
        url: str,
        input_bytes: Union[BytesIO, bytes],
        timeout_seconds: float = _DEFAULT_HTTP_TIMEOUT_UPLOAD_DATA_FRAME,
    ) -> None:
        """
        Upload bytes to an existing data frame.

        :param url: The url provided by Benchling for uploading to the data frame
        :param input_bytes: Data to upload as bytes or BytesIO
        :param timeout_seconds: Extends the normal HTTP timeout settings since DataFrame uploads can be large
            Use this to extend even further if streams are very large
        """
        # Use a completely different client instead of our configured self.client.httpx_client
        # Amazon will reject clients sending other headers besides the ones it expects
        httpx_response = httpx.put(
            url, headers=_aws_url_headers(), content=input_bytes, timeout=timeout_seconds
        )
        response = _response_from_httpx(httpx_response)
        raise_for_status(response)

    def upload_file(
        self, url: str, file: Path, timeout_seconds: float = _DEFAULT_HTTP_TIMEOUT_UPLOAD_DATA_FRAME
    ) -> None:
        """
        Upload a file to an existing data frame.

        :param url: The url provided by Benchling for uploading to the data frame
        :param file: A valid Path to an existing file containing the data to upload
        :param timeout_seconds: Extends the normal HTTP timeout settings since DataFrame uploads can be large
            Use this to extend even further if streams are very large
        """
        if file.is_dir():
            raise IsADirectoryError(
                f"Cannot write data frame from directory '{file}', specify a file instead"
            )
        # Use a completely different client instead of our configured self.client.httpx_client
        # Amazon will reject clients sending other headers besides the ones it expects
        files = {"file": open(file, "rb")}
        httpx_response = httpx.put(url, headers=_aws_url_headers(), files=files, timeout=timeout_seconds)
        response = _response_from_httpx(httpx_response)
        raise_for_status(response)

    @api_method
    def create_from_bytes(
        self,
        data_frame: DataFrameCreate,
        input_bytes: Union[BytesIO, bytes],
        timeout_seconds: float = _DEFAULT_HTTP_TIMEOUT_UPLOAD_DATA_FRAME,
    ) -> TaskHelper[DataFrame]:
        """
        Create a data frame from bytes or BytesIO data.

        :param data_frame: The DataFrameCreate specification for the data. This must be provided, as it cannot be inferred from file names.
        :param input_bytes: Data to upload as bytes or BytesIO
        :param timeout_seconds: Extends the normal HTTP timeout settings since DataFrame uploads can be large
            Use this to extend even further if streams are very large
        :return: A TaskHelper that can be polled to know when the data frame has completed processing
        :rtype: TaskHelper[DataFrame]
        """
        # This is a current limit of the DataFrame API. We may need additional methods in the future
        # to allow multi upload
        if not data_frame.manifest:
            raise InvalidDataFrameError("The data frame manifest must contain exactly 1 item")
        elif len(data_frame.manifest) != 1:
            raise InvalidDataFrameError(
                f"The data frame manifest contains {len(data_frame.manifest)} items. It must contain exactly 1"
            )
        created_data_frame = self.create(data_frame)
        manifest_item = created_data_frame.manifest[0]

        # This would be unexpected and probably an error from the API return. Likely not a user error. This check appeases MyPy.
        if manifest_item.url is None:
            raise InvalidDataFrameError(
                f"The data frame manifest URL is None. The data frame {created_data_frame.id} is not available for data upload."
            )
        self.upload_bytes(url=manifest_item.url, input_bytes=input_bytes, timeout_seconds=timeout_seconds)
        data_frame_update = DataFrameUpdate(upload_status=DataFrameUpdateUploadStatus.IN_PROGRESS)
        return self.update(data_frame_id=created_data_frame.id, data_frame=data_frame_update)

    @api_method
    def create_from_file(
        self,
        file: Path,
        data_frame: Optional[DataFrameCreate] = None,
        timeout_seconds: float = _DEFAULT_HTTP_TIMEOUT_UPLOAD_DATA_FRAME,
    ) -> TaskHelper[DataFrame]:
        """
        Create a data frame from file data.

        :param file: A valid Path to an existing file containing the data to upload
        :param data_frame: The DataFrameCreate specification for the data. If not provided, it will be inferred from the file name
        :param timeout_seconds: Extends the normal HTTP timeout settings since DataFrame uploads can be large
            Use this to extend even further if streams are very large
        :return: A TaskHelper that can be polled to know when the data frame has completed processing
        :rtype: TaskHelper[DataFrame]
        """
        if file.is_dir():
            raise IsADirectoryError(
                f"Cannot write data frame from directory '{file}', specify a file instead"
            )
        with open(file, "rb") as file_handle:
            input_bytes = file_handle.read()
        if not data_frame:
            data_frame = DataFrameCreate(
                name=f"{datetime.now()} {file.name}",
                manifest=[DataFrameCreateManifestManifestItem(file_name=file.name)],
            )
        return self.create_from_bytes(
            data_frame=data_frame, input_bytes=input_bytes, timeout_seconds=timeout_seconds
        )

    def download_data_frame_bytes(
        self, data_frame: DataFrame, timeout_seconds: float = _DEFAULT_HTTP_TIMEOUT_UPLOAD_DATA_FRAME
    ) -> List[BytesIO]:
        """
        Download data frame data to bytes.

        :param data_frame: The data frame to download
        :param timeout_seconds: Extends the normal HTTP timeout settings since DataFrame uploads can be large
            Use this to extend even further if streams are very large
        :return: An ordered list of BytesIO streams corresponding to a manifest item in the data frame
        :rtype: List[BytesIO]
        """
        if data_frame.upload_status != FileStatusUploadStatus.SUCCEEDED:
            raise DataFrameInProgressError(
                f"The data frame data cannot be downloaded until the status is {FileStatusUploadStatus.SUCCEEDED}. "
                f"The status of data frame {data_frame.id} is {data_frame.upload_status}"
            )
        data_frame_bytes = []
        for manifest_item in data_frame.manifest:
            # This should be present based on the status check above. Assertion satisfies MyPy
            assert (
                manifest_item.url is not None
            ), f"Unable to download data frame {data_frame.id}, URL was empty"
            with httpx.stream("GET", manifest_item.url, timeout=timeout_seconds) as download_stream:
                target_bytes = BytesIO()
                for chunk in download_stream.iter_bytes():
                    target_bytes.write(chunk)
                target_bytes.seek(0)
                data_frame_bytes.append(target_bytes)
        return data_frame_bytes

    def download_data_frame_files(
        self,
        data_frame: DataFrame,
        destination_path: Optional[Path] = None,
        timeout_seconds: float = _DEFAULT_HTTP_TIMEOUT_UPLOAD_DATA_FRAME,
    ) -> List[Path]:
        """
        Download data frame data to files.

        :param data_frame: The data frame to download
        :param destination_path: A target directory to place the files. File names will be created based on the manifest item file names.
            If not specified, a temp directory will be created. The caller is responsible for deleting this directory.
        :param timeout_seconds: Extends the normal HTTP timeout settings since DataFrame uploads can be large
            Use this to extend even further if streams are very large
        :return: An ordered list of downloaded file paths corresponding to a manifest item in the data frame
        :rtype: List[Path]
        """
        data_frame_files = []
        if not destination_path:
            destination_path = Path(tempfile.mkdtemp())
        elif destination_path.is_file():
            raise NotADirectoryError(
                f"The destination path '{destination_path}' is a file, specify a directory instead"
            )
        elif not destination_path.exists():
            raise NotADirectoryError(f"The destination path '{destination_path}' does not exist")
        if data_frame.upload_status != FileStatusUploadStatus.SUCCEEDED:
            raise DataFrameInProgressError(
                f"The data frame data cannot be downloaded until the status is {FileStatusUploadStatus.SUCCEEDED}. "
                f"The status of data frame {data_frame.id} is {data_frame.upload_status}"
            )
        for manifest_item in data_frame.manifest:
            target_path = destination_path / manifest_item.file_name
            data_frame_files.append(target_path)
            # This should be present based on the status check above. Assertion satisfies MyPy
            assert (
                manifest_item.url is not None
            ), f"Unable to download data frame {data_frame.id}, URL was empty"
            with open(target_path, "wb") as data_frame_handle:
                with httpx.stream("GET", manifest_item.url, timeout=timeout_seconds) as download_stream:
                    for chunk in download_stream.iter_bytes():
                        data_frame_handle.write(chunk)
        return data_frame_files

    @api_method
    def download_data_frame_bytes_by_id(
        self, data_frame_id: str, timeout_seconds: float = _DEFAULT_HTTP_TIMEOUT_UPLOAD_DATA_FRAME
    ) -> List[BytesIO]:
        """
        Download data frame data to files by data_frame_id.

        Fetches the data frame first, then downloads the files.

        :param data_frame_id: The id of the data frame to download
        :param timeout_seconds: Extends the normal HTTP timeout settings since DataFrame uploads can be large
            Use this to extend even further if streams are very large
        :return: An ordered list of BytesIO streams corresponding to a manifest item in the data frame
        :rtype: List[BytesIO]
        """
        data_frame = self.get_by_id(data_frame_id=data_frame_id)
        return self.download_data_frame_bytes(data_frame=data_frame, timeout_seconds=timeout_seconds)

    @api_method
    def download_data_frame_files_by_id(
        self,
        data_frame_id: str,
        destination_path: Optional[Path] = None,
        timeout_seconds: float = _DEFAULT_HTTP_TIMEOUT_UPLOAD_DATA_FRAME,
    ) -> List[Path]:
        """
        Download data frame data to files by data_frame_id.

        Fetches the data frame first, then downloads the files.

        :param data_frame_id: The id of the data frame to download
        :param destination_path: A target directory to place the files. File names will be created based on the manifest item file names.
            If not specified, a temp directory will be created. The caller is responsible for deleting this directory.
        :param timeout_seconds: Extends the normal HTTP timeout settings since DataFrame uploads can be large
            Use this to extend even further if streams are very large
        :return: An ordered list of downloaded file paths corresponding to a manifest item in the data frame
        :rtype: List[Path]
        """
        data_frame = self.get_by_id(data_frame_id=data_frame_id)
        return self.download_data_frame_files(
            data_frame=data_frame, destination_path=destination_path, timeout_seconds=timeout_seconds
        )


def _aws_url_headers() -> Dict[str, str]:
    return {"x-amz-server-side-encryption": "AES256"}


def _response_from_httpx(httpx_response: httpx.Response) -> Response:
    return Response(
        status_code=httpx_response.status_code,
        content=httpx_response.content,
        headers=httpx_response.headers,
        parsed=None,
    )



from typing import List, Optional

from benchling_api_client.v2.stable.api.schemas import (
    get_batch_schema,
    get_box_schema,
    get_container_schema,
    get_entity_schema,
    get_entry_schema,
    get_location_schema,
    get_plate_schema,
    get_request_schema,
    get_request_task_schema,
    get_result_schema,
    get_run_schema,
    get_workflow_task_schema,
    list_assay_result_schemas,
    list_assay_run_schemas,
    list_batch_schemas,
    list_box_schemas,
    list_container_schemas,
    list_entity_schemas,
    list_entry_schemas,
    list_location_schemas,
    list_plate_schemas,
    list_request_schemas,
    list_request_task_schemas,
    list_workflow_task_schemas,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import log_not_implemented
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.models import (
    AssayResultSchema,
    AssayResultSchemasPaginatedList,
    AssayRunSchema,
    AssayRunSchemasPaginatedList,
    BatchSchema,
    BatchSchemasPaginatedList,
    BoxSchema,
    BoxSchemasPaginatedList,
    ContainerSchema,
    ContainerSchemasPaginatedList,
    EntitySchema,
    EntitySchemasPaginatedList,
    EntrySchemaDetailed,
    EntrySchemasPaginatedList,
    LocationSchema,
    LocationSchemasPaginatedList,
    PlateSchema,
    PlateSchemasPaginatedList,
    RequestSchema,
    RequestSchemasPaginatedList,
    RequestTaskSchema,
    RequestTaskSchemasPaginatedList,
    WorkflowTaskSchema,
    WorkflowTaskSchemasPaginatedList,
)
from benchling_sdk.services.v2.base_service import BaseService


class SchemaService(BaseService):
    """
    Schemas.

    Schemas represent custom configuration of objects in Benchling. See https://docs.benchling.com/docs/schemas in
    our documentation on how Schemas impact our developers.

    See API docs at https://benchling.com/api/reference#/Schemas
    """

    @api_method
    def get_assay_result_schema_by_id(self, schema_id: str) -> AssayResultSchema:
        """
        Get a Result schema by ID.

        See https://benchling.com/api/reference#/Schemas/getResultSchema
        """
        response = get_result_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _assay_result_schemas_page(
        self,
        *,
        modified_at: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[AssayResultSchemasPaginatedList]:
        """
        Get a Result schema by ID.

        See https://benchling.com/api/reference#/Schemas/getResultSchema
        """
        response = list_assay_result_schemas.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_assay_result_schemas(
        self,
        *,
        modified_at: Optional[str] = None,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[AssayResultSchema]:
        """
        List assay result schemas.

        See https://benchling.com/api/reference#/Schemas/listAssayResultSchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[AssayResultSchemasPaginatedList]:
            return self._assay_result_schemas_page(
                modified_at=modified_at,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: AssayResultSchemasPaginatedList) -> Optional[List[AssayResultSchema]]:
            return body.assay_result_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_assay_run_schema_by_id(self, schema_id: str) -> AssayRunSchema:
        """
        Get a Run schema by ID.

        See https://benchling.com/api/reference#/Schemas/getRunSchema
        """
        response = get_run_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _assay_run_schemas_page(
        self,
        *,
        modified_at: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[AssayRunSchemasPaginatedList]:
        response = list_assay_run_schemas.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_assay_run_schemas(
        self,
        *,
        modified_at: Optional[str] = None,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[AssayRunSchema]:
        """
        List assay run schemas.

        See https://benchling.com/api/reference#/Schemas/listAssayRunSchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[AssayRunSchemasPaginatedList]:
            return self._assay_run_schemas_page(
                modified_at=modified_at,
                next_token=next_token,
                page_size=page_size,
            )

        def runs_extractor(body: AssayRunSchemasPaginatedList) -> Optional[List[AssayRunSchema]]:
            return body.assay_run_schemas

        return PageIterator(api_call, runs_extractor)

    @api_method
    def get_batch_schemas_by_id(self, schema_id: str) -> BatchSchema:
        """
        Get a batch schema by ID.

        See https://benchling.com/api/reference#/Schemas/getBatchSchema
        """
        response = get_batch_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _batch_schemas_page(
        self,
        *,
        modified_at: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[BatchSchemasPaginatedList]:
        response = list_batch_schemas.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_batch_schemas(
        self,
        *,
        modified_at: Optional[str] = None,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[BatchSchema]:
        """
        List batch schemas.

        See https://benchling.com/api/reference#/Schemas/listBatchSchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[BatchSchemasPaginatedList]:
            return self._batch_schemas_page(
                modified_at=modified_at,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: BatchSchemasPaginatedList) -> Optional[List[BatchSchema]]:
            return body.batch_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_box_schema_by_id(self, schema_id: str) -> BoxSchema:
        """
        Get a box schema by ID.

        See https://benchling.com/api/reference#/Schemas/getBoxSchema
        """
        response = get_box_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _box_schemas_page(
        self,
        *,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[BoxSchemasPaginatedList]:
        response = list_box_schemas.sync_detailed(
            client=self.client,
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_box_schemas(
        self,
        *,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[BoxSchema]:
        """
        List box schemas.

        See https://benchling.com/api/reference#/Schemas/listBoxSchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[BoxSchemasPaginatedList]:
            return self._box_schemas_page(
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: BoxSchemasPaginatedList) -> Optional[List[BoxSchema]]:
            return body.box_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_container_schema_by_id(self, schema_id: str) -> ContainerSchema:
        """
        Get a container schema by ID.

        See https://benchling.com/api/reference#/Schemas/getContainerSchema
        """
        response = get_container_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _container_schemas_page(
        self,
        *,
        modified_at: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[ContainerSchemasPaginatedList]:
        response = list_container_schemas.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_container_schemas(
        self,
        *,
        modified_at: Optional[str] = None,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[ContainerSchema]:
        """
        List container schemas.

        See https://benchling.com/api/reference#/Schemas/listContainerSchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[ContainerSchemasPaginatedList]:
            return self._container_schemas_page(
                modified_at=modified_at,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: ContainerSchemasPaginatedList) -> List[ContainerSchema]:
            return body.container_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_entity_schema_by_id(self, schema_id: str) -> EntitySchema:
        """
        Get an entity schema by ID.

        See https://benchling.com/api/reference#/Schemas/getEntitySchema
        """
        response = get_entity_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _entity_schemas_page(
        self,
        *,
        modified_at: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[EntitySchemasPaginatedList]:
        response = list_entity_schemas.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_entity_schemas(
        self,
        *,
        modified_at: Optional[str] = None,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[EntitySchema]:
        """
        List entity schemas.

        See https://benchling.com/api/reference#/Schemas/listEntitySchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[EntitySchemasPaginatedList]:
            return self._entity_schemas_page(
                modified_at=modified_at,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: EntitySchemasPaginatedList) -> Optional[List[EntitySchema]]:
            return body.entity_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_entry_schema_by_id(self, schema_id: str) -> EntrySchemaDetailed:
        """
        Get an Entry schema by ID.

        See https://benchling.com/api/reference#/Schemas/getEntrySchema
        """
        response = get_entry_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _entry_schemas_page(
        self,
        *,
        modified_at: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[EntrySchemasPaginatedList]:
        response = list_entry_schemas.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_entry_schemas(
        self,
        *,
        modified_at: Optional[str] = None,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[EntrySchemaDetailed]:
        """
        List Entry schemas.

        See https://benchling.com/api/reference#/Schemas/listEntrySchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[EntrySchemasPaginatedList]:
            return self._entry_schemas_page(
                modified_at=modified_at,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: EntrySchemasPaginatedList) -> Optional[List[EntrySchemaDetailed]]:
            return body.entry_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_location_schema_by_id(self, schema_id: str) -> LocationSchema:
        """
        Get a location schema by ID.

        See https://benchling.com/api/reference#/Schemas/getLocationSchema
        """
        response = get_location_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _location_schemas_page(
        self,
        *,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[LocationSchemasPaginatedList]:
        response = list_location_schemas.sync_detailed(
            client=self.client,
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_location_schemas(
        self,
        *,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[LocationSchema]:
        """
        List location schemas.

        See https://benchling.com/api/reference#/Schemas/listLocationSchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[LocationSchemasPaginatedList]:
            return self._location_schemas_page(
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: LocationSchemasPaginatedList) -> List[LocationSchema]:
            return body.location_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_plate_schema_by_id(self, schema_id: str) -> PlateSchema:
        """
        Get a plate schema by ID.

        See https://benchling.com/api/reference#/Schemas/getPlateSchema
        """
        response = get_plate_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _plate_schemas_page(
        self,
        *,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[PlateSchemasPaginatedList]:
        response = list_plate_schemas.sync_detailed(
            client=self.client,
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_plate_schemas(
        self,
        *,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[PlateSchema]:
        """
        List plate schemas.

        See https://benchling.com/api/reference#/Schemas/listPlateSchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[PlateSchemasPaginatedList]:
            return self._plate_schemas_page(
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: PlateSchemasPaginatedList) -> Optional[List[PlateSchema]]:
            return body.plate_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_request_schema_by_id(self, schema_id: str) -> RequestSchema:
        """
        Get a Request schema by ID.

        See https://benchling.com/api/reference#/Schemas/getRequestSchema
        """
        response = get_request_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _request_schemas_page(
        self,
        *,
        modified_at: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[RequestSchemasPaginatedList]:
        response = list_request_schemas.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_request_schemas(
        self,
        *,
        registry_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[RequestSchema]:
        """
        List Request schemas.

        See https://benchling.com/api/reference#/Schemas/listRequestSchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(next_token: NextToken) -> Response[RequestSchemasPaginatedList]:
            return self._request_schemas_page(
                modified_at=modified_at,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: RequestSchemasPaginatedList) -> Optional[List[RequestSchema]]:
            return body.request_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_request_task_schema_by_id(self, schema_id: str) -> RequestTaskSchema:
        """
        Get a Request Task schema by ID.

        See https://benchling.com/api/reference#/Schemas/getRequestTaskSchema
        """
        response = get_request_task_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _request_task_schemas_page(
        self,
        *,
        next_token: Optional[str] = None,
        modified_at: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[RequestTaskSchemasPaginatedList]:
        response = list_request_task_schemas.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_request_task_schemas(
        self,
        *,
        modified_at: Optional[str] = None,
        registry_id: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[RequestTaskSchema]:
        """
        List Request Task schemas.

        See https://benchling.com/api/reference#/Schemas/listRequestTaskSchemas
        """
        if registry_id:
            log_not_implemented("registry_id")

        def api_call(
            next_token: NextToken,
        ) -> Response[RequestTaskSchemasPaginatedList]:
            return self._request_task_schemas_page(
                modified_at=modified_at, next_token=next_token, page_size=page_size
            )

        def results_extractor(body: RequestTaskSchemasPaginatedList) -> Optional[List[RequestTaskSchema]]:
            return body.request_task_schemas

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_workflow_task_schema_by_id(self, schema_id: str) -> WorkflowTaskSchema:
        """
        Get a Workflow Task schema by ID.

        See https://benchling.com/api/reference#/Schemas/getWorkflowTaskSchema
        """
        response = get_workflow_task_schema.sync_detailed(client=self.client, schema_id=schema_id)
        return model_from_detailed(response)

    @api_method
    def _workflow_task_schemas_page(
        self,
        *,
        modified_at: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[WorkflowTaskSchemasPaginatedList]:
        response = list_workflow_task_schemas.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_workflow_task_schemas(
        self, *, modified_at: Optional[str] = None, page_size: Optional[int] = 50
    ) -> PageIterator[WorkflowTaskSchema]:
        """
        List Workflow Task schemas.

        See https://benchling.com/api/reference#/Schemas/listWorkflowTaskSchemas
        """

        def api_call(
            next_token: NextToken,
        ) -> Response[WorkflowTaskSchemasPaginatedList]:
            return self._workflow_task_schemas_page(
                modified_at=modified_at, next_token=next_token, page_size=page_size
            )

        def results_extractor(body: WorkflowTaskSchemasPaginatedList) -> Optional[List[WorkflowTaskSchema]]:
            return body.workflow_task_schemas

        return PageIterator(api_call, results_extractor)



import base64
import hashlib


def encode_base64(input: bytes, charset: str = "utf-8") -> str:
    """Encode bytes into Base64."""
    file_bytes = base64.encodebytes(input)
    return str(file_bytes, charset)


def calculate_md5(input: bytes) -> str:
    """Calculate an MD5 hash from bytes."""
    return hashlib.md5(input).hexdigest()



from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.locations import (
    archive_locations,
    bulk_get_locations,
    create_location,
    get_location,
    list_locations,
    unarchive_locations,
    update_location,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    ListLocationsSort,
    Location,
    LocationCreate,
    LocationsArchivalChange,
    LocationsArchive,
    LocationsArchiveReason,
    LocationsPaginatedList,
    LocationsUnarchive,
    LocationUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class LocationService(BaseService):
    """
    Locations.

    Manage locations objects. Like all storage, every Location has a barcode that is unique across the registry.

    See https://benchling.com/api/reference#/Locations
    """

    @api_method
    def get_by_id(self, location_id: str) -> Location:
        """
        Get a location by ID.

        See https://benchling.com/api/reference#/Locations/getLocation
        """
        response = get_location.sync_detailed(client=self.client, location_id=location_id)
        return model_from_detailed(response)

    @api_method
    def _locations_page(
        self,
        *,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListLocationsSort] = None,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ancestor_storage_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
    ) -> Response[LocationsPaginatedList]:
        response = list_locations.sync_detailed(
            client=self.client,
            sort=none_as_unset(sort),
            schema_id=none_as_unset(schema_id),
            modified_at=none_as_unset(modified_at),
            created_at=none_as_unset(created_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            ancestor_storage_id=none_as_unset(ancestor_storage_id),
            archive_reason=none_as_unset(archive_reason),
            ids=none_as_unset(optional_array_query_param(ids)),
            barcodes=none_as_unset(optional_array_query_param(barcodes)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        sort: Optional[Union[str, ListLocationsSort]] = None,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ancestor_storage_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[Location]:
        """
        List locations.

        See https://benchling.com/api/reference#/Locations/listLocations
        """

        def api_call(next_token: NextToken) -> Response[LocationsPaginatedList]:
            return self._locations_page(
                sort=_translate_to_string_enum(ListLocationsSort, sort),
                schema_id=schema_id,
                modified_at=modified_at,
                created_at=created_at,
                name=name,
                name_includes=name_includes,
                ancestor_storage_id=ancestor_storage_id,
                archive_reason=archive_reason,
                ids=ids,
                barcodes=barcodes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                creator_ids=creator_ids,
                schema_fields=schema_fields,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: LocationsPaginatedList) -> Optional[List[Location]]:
            return body.locations

        return PageIterator(api_call, results_extractor)

    @api_method
    def bulk_get(
        self, *, location_ids: Optional[Iterable[str]] = None, barcodes: Optional[Iterable[str]] = None
    ) -> Optional[List[Location]]:
        """
        Bulk get locations.

        See https://benchling.com/api/reference#/Locations/bulkGetLocations
        """
        location_id_string = optional_array_query_param(location_ids)
        barcode_string = optional_array_query_param(barcodes)
        response = bulk_get_locations.sync_detailed(
            client=self.client,
            location_ids=none_as_unset(location_id_string),
            barcodes=none_as_unset(barcode_string),
        )
        locations_list = model_from_detailed(response)
        return locations_list.locations

    @api_method
    def create(self, location: LocationCreate) -> Location:
        """
        Create a location.

        See https://benchling.com/api/reference#/Locations/createLocation
        """
        response = create_location.sync_detailed(client=self.client, json_body=location)
        return model_from_detailed(response)

    @api_method
    def update(self, location_id: str, location: LocationUpdate) -> Location:
        """
        Update a location.

        See https://benchling.com/api/reference#/Locations/updateLocation
        """
        response = update_location.sync_detailed(
            client=self.client, location_id=location_id, json_body=location
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self, location_ids: Iterable[str], reason: LocationsArchiveReason, should_remove_barcodes: bool
    ) -> LocationsArchivalChange:
        """
        Archive locations.

        See https://benchling.com/api/reference#/Locations/archiveLocations
        """
        archive_request = LocationsArchive(
            location_ids=list(location_ids), reason=reason, should_remove_barcodes=should_remove_barcodes
        )
        response = archive_locations.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, location_ids: Iterable[str]) -> LocationsArchivalChange:
        """
        Unarchive locations.

        See https://benchling.com/api/reference#/Locations/unarchiveLocations
        """
        unarchive_request = LocationsUnarchive(location_ids=list(location_ids))
        response = unarchive_locations.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)



from typing import Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.entries import (
    archive_entries,
    bulk_get_entries,
    create_entry,
    get_entry,
    get_entry_template,
    get_external_file_metadata,
    list_entries,
    list_entry_templates,
    unarchive_entries,
    update_entry,
    update_entry_template,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    EntriesArchivalChange,
    EntriesArchive,
    EntriesArchiveReason,
    EntriesPaginatedList,
    EntriesUnarchive,
    Entry,
    EntryCreate,
    EntryExternalFile,
    EntryTemplate,
    EntryTemplatesPaginatedList,
    EntryTemplateUpdate,
    EntryUpdate,
    ListEntriesReviewStatus,
    ListEntriesSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class EntryService(BaseService):
    """
    Entries.

    Entries are rich text documents that allow you to capture all of your experimental data in one place.

    See https://benchling.com/api/reference#/Entries
    """

    @api_method
    def get_entry_by_id(self, entry_id: str, returning: Optional[Iterable[str]] = None) -> Entry:
        """
        Get a notebook entry by ID.

        See https://benchling.com/api/reference#/Entries/getEntry
        """
        returning_string = optional_array_query_param(returning)
        response = get_entry.sync_detailed(
            client=self.client, entry_id=entry_id, returning=none_as_unset(returning_string)
        )
        wrapped_entry = model_from_detailed(response)
        return wrapped_entry.entry

    @api_method
    def _entries_page(
        self,
        *,
        sort: Optional[ListEntriesSort] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        review_status: Optional[ListEntriesReviewStatus] = None,
        mentioned_in: Optional[str] = None,
        mentions: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        display_ids: Optional[Iterable[str]] = None,
        assigned_reviewer_ids_any_of: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
        author_idsany_of: Optional[Iterable[str]] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> Response[EntriesPaginatedList]:
        response = list_entries.sync_detailed(
            client=self.client,
            sort=none_as_unset(sort),
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            project_id=none_as_unset(project_id),
            archive_reason=none_as_unset(archive_reason),
            review_status=none_as_unset(review_status),
            mentioned_in=none_as_unset(mentioned_in),
            mentions=none_as_unset(mentions),
            ids=none_as_unset(optional_array_query_param(ids)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            display_ids=none_as_unset(optional_array_query_param(display_ids)),
            assigned_reviewer_idsany_of=none_as_unset(
                optional_array_query_param(assigned_reviewer_ids_any_of)
            ),
            schema_id=none_as_unset(schema_id),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
            author_idsany_of=none_as_unset(optional_array_query_param(author_idsany_of)),
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_entries(
        self,
        *,
        sort: Optional[Union[str, ListEntriesSort]] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        review_status: Optional[ListEntriesReviewStatus] = None,
        mentioned_in: Optional[str] = None,
        mentions: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        display_ids: Optional[Iterable[str]] = None,
        assigned_reviewer_ids_any_of: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        page_size: Optional[int] = 50,
        author_idsany_of: Optional[Iterable[str]] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> PageIterator[Entry]:
        """
        List notebook entries.

        See https://benchling.com/api/reference#/Entries/listEntries
        """

        def api_call(next_token: NextToken) -> Response[EntriesPaginatedList]:
            return self._entries_page(
                sort=_translate_to_string_enum(ListEntriesSort, sort),
                modified_at=modified_at,
                name=name,
                project_id=project_id,
                archive_reason=archive_reason,
                review_status=review_status,
                mentioned_in=mentioned_in,
                mentions=mentions,
                ids=ids,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                creator_ids=creator_ids,
                display_ids=display_ids,
                assigned_reviewer_ids_any_of=assigned_reviewer_ids_any_of,
                schema_id=schema_id,
                next_token=next_token,
                page_size=page_size,
                author_idsany_of=author_idsany_of,
                returning=returning,
            )

        def results_extractor(body: EntriesPaginatedList) -> Optional[List[Entry]]:
            return body.entries

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_external_file(self, entry_id: str, external_file_id: str) -> EntryExternalFile:
        """
        Retrieve the metadata for an external file.

        Use the `download_url` to download the actual file.

        See https://benchling.com/api/reference#/Entries/getExternalFileMetadata
        """
        response = get_external_file_metadata.sync_detailed(
            client=self.client, entry_id=entry_id, external_file_id=external_file_id
        )
        wrapped_file = model_from_detailed(response)
        return wrapped_file.external_file

    @api_method
    def create_entry(self, entry: EntryCreate) -> Entry:
        """
        Create a notebook entry.

        See https://benchling.com/api/reference#/Entries/createEntry
        """
        response = create_entry.sync_detailed(client=self.client, json_body=entry)
        return model_from_detailed(response)

    @api_method
    def update_entry(
        self, entry_id: str, entry: EntryUpdate, returning: Optional[Iterable[str]] = None
    ) -> Entry:
        """
        Update a notebook entry's metadata.

        See https://benchling.com/api/reference#/Entries/updateEntry
        """
        returning_string = optional_array_query_param(returning)
        response = update_entry.sync_detailed(
            client=self.client, entry_id=entry_id, json_body=entry, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)

    @api_method
    def bulk_get_entries(
        self,
        entry_ids: Optional[Iterable[str]] = None,
        display_ids: Optional[Iterable[str]] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> Optional[List[Entry]]:
        """
        Bulk get notebook entries.

        See https://benchling.com/api/reference#/Entries/bulkGetEntries
        """
        entry_id_string = optional_array_query_param(entry_ids)
        display_id_string = optional_array_query_param(display_ids)
        returning_string = optional_array_query_param(returning)
        response = bulk_get_entries.sync_detailed(
            client=self.client,
            entry_ids=none_as_unset(entry_id_string),
            display_ids=none_as_unset(display_id_string),
            returning=none_as_unset(returning_string),
        )
        entries = model_from_detailed(response)
        return entries.entries

    @api_method
    def archive_entries(
        self, entry_ids: Iterable[str], reason: EntriesArchiveReason
    ) -> EntriesArchivalChange:
        """
        Archive notebook entries.

        See https://benchling.com/api/reference#/Entries/archiveEntries
        """
        archive_request = EntriesArchive(entry_ids=list(entry_ids), reason=reason)
        response = archive_entries.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive_entries(self, entry_ids: Iterable[str]) -> EntriesArchivalChange:
        """
        Unarchive notebook entries.

        See https://benchling.com/api/reference#/Entries/unarchiveEntries
        """
        unarchive_request = EntriesUnarchive(entry_ids=list(entry_ids))
        response = unarchive_entries.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def _entry_templates_page(
        self,
        *,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        template_collection_id: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
        returning: Optional[Iterable[str]] = None,
    ) -> Response[EntryTemplatesPaginatedList]:
        response = list_entry_templates.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            template_collection_id=none_as_unset(template_collection_id),
            ids=none_as_unset(optional_array_query_param(ids)),
            schema_id=none_as_unset(schema_id),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_entry_templates(
        self,
        *,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        template_collection_id: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        schema_id: Optional[str] = None,
        page_size: Optional[int] = 50,
        returning: Optional[Iterable[str]] = None,
    ) -> PageIterator[EntryTemplate]:
        """
        List entry templates.

        See https://benchling.com/api/reference#/Entries/listEntryTemplates
        """

        def api_call(next_token: NextToken) -> Response[EntryTemplatesPaginatedList]:
            return self._entry_templates_page(
                modified_at=modified_at,
                name=name,
                template_collection_id=template_collection_id,
                ids=ids,
                schema_id=schema_id,
                next_token=next_token,
                page_size=page_size,
                returning=returning,
            )

        def results_extractor(body: EntryTemplatesPaginatedList) -> Optional[List[EntryTemplate]]:
            return body.entry_templates

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_entry_template_by_id(
        self, entry_template_id: str, returning: Optional[Iterable[str]] = None
    ) -> EntryTemplate:
        """
        Get a notebook template entry by ID.

        See https://benchling.com/api/reference#/Entries/getTemplateEntry
        """
        returning_string = optional_array_query_param(returning)
        # TODO rename get_entry_template
        response = get_entry_template.sync_detailed(
            client=self.client, entry_template_id=entry_template_id, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)

    @api_method
    def update_entry_template(
        self,
        entry_template_id: str,
        entry_template: EntryTemplateUpdate,
        returning: Optional[Iterable[str]] = None,
    ) -> EntryTemplate:
        """
        Update a notebook entry template's metadata.

        See https://benchling.com/api/reference#/Entries/updateEntryTemplate
        """
        returning_string = optional_array_query_param(returning)
        response = update_entry_template.sync_detailed(
            client=self.client,
            entry_template_id=entry_template_id,
            json_body=entry_template,
            returning=none_as_unset(returning_string),
        )
        return model_from_detailed(response)



from __future__ import annotations

from functools import cached_property
from typing import Optional, TYPE_CHECKING

from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.logging_helpers import log_stability_warning, StabilityLevel
from benchling_sdk.helpers.retry_helpers import RetryStrategy
from benchling_sdk.services.v2.base_service import BaseService

if TYPE_CHECKING:
    from benchling_sdk.services.v2.v2_alpha_service import V2AlphaService
    from benchling_sdk.services.v2.v2_beta_service import V2BetaService
    from benchling_sdk.services.v2.v2_stable_service import V2StableService


class V2Service(BaseService):
    """
    V2.

    Namespace containing support for the V2 Benchling API.
    """

    _v2_alpha_service: Optional[V2AlphaService]
    _v2_beta_service: Optional[V2BetaService]

    def __init__(self, client: Client, retry_strategy: Optional[RetryStrategy] = None):
        """
        Initialize a service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        super().__init__(client, retry_strategy)
        self._v2_alpha_service = None
        self._v2_beta_service = None

    # Note that alpha & beta do not use the @cached_property pattern that we've used for all other
    # service accessors; instead, we've rolled our own equivalent caching logic using an attribute.
    # That's because we want some additional behavior - logging a warning message - to happen on
    # every call to those property methods, not just on the first call. We could consider changing
    # this in a future release so it only logs once, in which case we can use @cached_property.

    @property
    def alpha(self) -> V2AlphaService:
        """
        V2-alpha.

        Alpha endpoints have different stability guidelines than other stable endpoints.

        See https://benchling.com/api/v2-alpha/reference
        """
        log_stability_warning(StabilityLevel.ALPHA, package="v2.alpha")

        if self._v2_alpha_service is None:
            from .v2.v2_alpha_service import V2AlphaService

            self._v2_alpha_service = self._create_service(V2AlphaService)

        return self._v2_alpha_service

    @property
    def beta(self) -> V2BetaService:
        """
        V2-beta.

        Beta endpoints have different stability guidelines than other stable endpoints.

        See https://benchling.com/api/v2-beta/reference
        """
        log_stability_warning(StabilityLevel.BETA, package="v2.beta")
        if self._v2_beta_service is None:
            from .v2.v2_beta_service import V2BetaService

            self._v2_beta_service = self._create_service(V2BetaService)

        return self._v2_beta_service

    @cached_property
    def stable(self) -> V2StableService:
        """
        Stable.

        See https://docs.benchling.com/docs/stability#are-breaking-changes-made-to-the-api-or-events
        """
        from .v2.v2_stable_service import V2StableService

        return self._create_service(V2StableService)



from __future__ import annotations

import base64
from datetime import datetime, timedelta, timezone
from json import JSONDecodeError
import threading
import typing
from typing import NoReturn, Optional
from urllib.parse import urljoin

from benchling_api_client.v2.benchling_client import AuthorizationMethod, BenchlingApiClient
import httpx

from benchling_sdk.errors import BenchlingError
from benchling_sdk.helpers.logging_helpers import sdk_logger

MINIMUM_TOKEN_EXPIRY_BUFFER = 60


class Token:
    """Represents an OAuth2 token response model."""

    def __init__(self, access_token: str, refresh_time: datetime):
        """
        Initialize Token.

        :param access_token: The raw token value for authorizing with the API
        :param refresh_time: Calculated value off of token time-to-live for when a new token should be generated.
        """
        self.access_token = access_token
        self.refresh_time = refresh_time

    def valid(self) -> bool:
        """Return whether token is still valid for use or should be regenerated."""
        return datetime.now(timezone.utc) < self.refresh_time

    @classmethod
    def from_token_response(cls, token_response) -> Token:
        """
        Construct Token from deserializing token endpoint response.

        Deserializes response from token endpoint and calculates expiry time with buffer for when token should be
        regenerated.

        :param token_response: The response from an RFC6749 POST /token endpoint.
        """
        token_type: str = token_response.get("token_type")
        access_token: str = token_response.get("access_token")
        expires_in: float = token_response.get("expires_in")
        assert token_type == "Bearer"
        # Add in a buffer to safeguard against race conditions with token expiration.
        # Buffer is 10% of expires_in time, clamped between [1, MINIMUM_TOKEN_EXPIRY_BUFFER] seconds.
        refresh_delta = expires_in - max(1, min(MINIMUM_TOKEN_EXPIRY_BUFFER, expires_in * 0.1))
        refresh_time = datetime.now(timezone.utc) + timedelta(seconds=refresh_delta)
        return cls(access_token, refresh_time)


class ClientCredentialsOAuth2(AuthorizationMethod):
    """
    OAuth2 client credentials for authorization.

    Use in combination with the Benchling() client constructor to be authorized with OAuth2 client_credentials grant
    type.
    """

    _data_for_token_request: typing.ClassVar[dict] = {
        "grant_type": "client_credentials",
    }

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        token_url: Optional[str] = None,
        httpx_client: Optional[httpx.Client] = None,
    ):
        """
        Initialize ClientCredentialsOAuth2.

        :param client_id: Client id in client_credentials grant type
        :param client_secret: Client secret in client_credentials grant type
        :param token_url: A fully-qualified URL pointing at the access token request endpoint such as
                          https://benchling.com/api/v2/token. Can be omitted to default to /api/v2/token appended to
                          the server base URL.
        :param httpx_client: An optional httpx Client which will be used to execute HTTP calls. The Client can be used
                             to modify the behavior of the HTTP calls made to Benchling through means such as adding
                             proxies and certificates or introducing retry logic for transport-level errors.
        """
        self._token_url = token_url
        token_encoded = base64.b64encode(f"{client_id}:{client_secret}".encode())
        self._header_for_token_request = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": f"Basic {token_encoded.decode()}",
            "User-Agent": BenchlingApiClient._get_user_agent("BenchlingSDK", "benchling-sdk"),
        }
        self._token: Optional[Token] = None
        self._lock = threading.Lock()
        if not httpx_client:
            httpx_client = httpx.Client()
        self.httpx_client = httpx_client

    def vend_new_token(self, base_url: str):
        """Make RFC6749 request to token URL to generate a new bearer token for client credentials OAuth2 flow."""
        token_url = self._token_url if self._token_url is not None else urljoin(base_url, "/api/v2/token")
        response: httpx.Response = self.httpx_client.post(
            token_url,
            data=ClientCredentialsOAuth2._data_for_token_request,
            headers=self._header_for_token_request,
        )

        if response.status_code == 200:
            as_json = response.json()
            self._token = Token.from_token_response(as_json)
        else:
            _raise_error_from_response(response)

    def get_authorization_header(self, base_url: str) -> str:
        """
        Generate HTTP Authorization request header.

        If a token has not yet been requested or is close to its expiry time, a new token is requested.
        Otherwise, re-use existing valid token.
        """
        with self._lock:
            if self._token is None or not self._token.valid():
                self.vend_new_token(base_url)
        assert self._token is not None
        return f"Bearer {self._token.access_token}"


def _raise_error_from_response(response: httpx.Response) -> NoReturn:
    json_content = None
    # Rather than rely on Content-Type header, try to parse JSON
    # If the response isn't JSON, just swallow the exception
    try:
        json_content = response.json()
    except JSONDecodeError as e:
        sdk_logger.debug("Received error response without JSON OAuth vending token", e)
    raise BenchlingError(
        status_code=response.status_code,
        headers=response.headers,
        json=json_content,
        content=response.content,
        parsed=None,
    )



import datetime
from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.workflow_flowcharts import (
    get_workflow_flowchart,
    list_workflow_flowcharts,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import ListWorkflowFlowchartsSort, WorkflowFlowchart, WorkflowFlowchartPaginatedList
from benchling_sdk.services.v2.base_service import BaseService


class WorkflowFlowchartService(BaseService):
    """
    Workflow Flowcharts.

    Workflow flowcharts represent the nodes and edges that a flowchart is comprised of.

    See https://benchling.com/api/reference#/Workflow%20Flowcharts
    """

    @api_method
    def get_by_id(self, workflow_flowchart_id: str) -> WorkflowFlowchart:
        """
        Get a workflow flowchart.

        See https://benchling.com/api/reference#/Workflow%20Flowcharts/getWorkflowFlowchart
        """
        response = get_workflow_flowchart.sync_detailed(
            client=self.client, workflow_flowchart_id=workflow_flowchart_id
        )
        return model_from_detailed(response)

    @api_method
    def _workflow_tasks_page(
        self,
        sort: Optional[ListWorkflowFlowchartsSort] = None,
        ids: Optional[Iterable[str]] = None,
        created_at: Optional[datetime.date] = None,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
    ) -> Response[WorkflowFlowchartPaginatedList]:
        response = list_workflow_flowcharts.sync_detailed(
            client=self.client,
            ids=none_as_unset(optional_array_query_param(ids)),
            created_at=none_as_unset(created_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
            sort=none_as_unset(sort),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        sort: Optional[ListWorkflowFlowchartsSort] = None,
        ids: Optional[Iterable[str]] = None,
        created_at: Optional[datetime.date] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[WorkflowFlowchart]:
        """
        List workflow flowcharts.

        See https://benchling.com/api/reference#/Workflow%20Tasks/listWorkflowFlowcharts
        """

        def api_call(next_token: NextToken) -> Response[WorkflowFlowchartPaginatedList]:
            return self._workflow_tasks_page(
                sort=sort,
                ids=ids,
                created_at=created_at,
                page_size=page_size,
                next_token=next_token,
            )

        def results_extractor(body: WorkflowFlowchartPaginatedList) -> Optional[List[WorkflowFlowchart]]:
            return body.workflow_flowcharts

        return PageIterator(api_call, results_extractor)



from typing import Iterable, Optional

from benchling_api_client.v2.beta.api.entries import update_entry
from benchling_api_client.v2.beta.models.entry import Entry
from benchling_api_client.v2.beta.models.entry_update import EntryUpdate

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaEntryService(BaseService):
    """
    V2-Beta Entries.

    Entries are rich text documents that allow you to capture all of your experimental data in one place.

    https://benchling.com/api/v2-beta/reference#/Entries
    """

    @api_method
    def update_entry(
        self, entry_id: str, entry: EntryUpdate, returning: Optional[Iterable[str]] = None
    ) -> Entry:
        """
        Update a notebook entry's metadata.

        See https://benchling.com/api/v2-beta/reference#/Entries/updateEntry
        """
        returning_string = optional_array_query_param(returning)
        response = update_entry.sync_detailed(
            client=self.client, entry_id=entry_id, json_body=entry, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)



from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.containers import (
    archive_containers,
    bulk_create_containers,
    bulk_get_containers,
    bulk_update_containers,
    checkin_containers,
    checkout_containers,
    create_container,
    delete_container_content,
    get_container,
    get_container_content,
    list_container_contents,
    list_containers,
    print_labels,
    reserve_containers,
    transfer_into_container,
    transfer_into_containers,
    unarchive_containers,
    update_container,
    update_container_content,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import (
    BulkCreateContainersAsyncTaskResponse,
    BulkUpdateContainersAsyncTaskResponse,
    Container,
    ContainerBulkUpdateItem,
    ContainerContent,
    ContainerContentUpdate,
    ContainerCreate,
    ContainersArchivalChange,
    ContainersArchive,
    ContainersArchiveReason,
    ContainersBulkCreateRequest,
    ContainersBulkUpdateRequest,
    ContainersCheckin,
    ContainersCheckout,
    ContainersPaginatedList,
    ContainersUnarchive,
    ContainerTransfer,
    ContainerUpdate,
    ListContainersCheckoutStatus,
    ListContainersSort,
    Measurement,
    MultipleContainersTransfer,
    MultipleContainersTransfersList,
    PrintLabels,
    SampleRestrictionStatus,
    TransfersAsyncTaskResponse,
)
from benchling_sdk.services.v2.base_service import BaseService


class ContainerService(BaseService):
    """
    Containers.

    Containers are the backbone of sample management in Benchling. They represent physical containers, such as
    tubes or wells, that hold quantities of biological samples (represented by the batches inside the container).
    The container itself tracks its total volume, and the concentration of every batch inside of it.

    Containers are all associated with schemas, which define the type of the container (e.g. "Tube") along with the
    fields that are tracked.

    Like all storage, every container has a barcode that is unique across the registry.

    See https://benchling.com/api/reference#/Containers
    """

    @api_method
    def get_by_id(self, container_id: str, returning: Optional[Iterable[str]] = None) -> Container:
        """
        Get a container by ID.

        See https://benchling.com/api/reference#/Containers/getContainer
        """
        returning_string = optional_array_query_param(returning)
        response = get_container.sync_detailed(
            client=self.client, container_id=container_id, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)

    @api_method
    def _containers_page(
        self,
        *,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListContainersSort] = None,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ancestor_storage_id: Optional[str] = None,
        storage_contents_id: Optional[str] = None,
        storage_contents_ids: Optional[List[str]] = None,
        checkout_status: Optional[ListContainersCheckoutStatus] = None,
        ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        checkout_assignee_ids_any_of: Optional[Iterable[str]] = None,
        restricted_sample_party_ids_all_of: Optional[Iterable[str]] = None,
        restricted_sample_party_ids_any_of: Optional[Iterable[str]] = None,
        restricted_sample_party_ids_none_of: Optional[Iterable[str]] = None,
        restriction_status: Optional[SampleRestrictionStatus] = None,
        sample_owner_ids_all_of: Optional[Iterable[str]] = None,
        sample_owner_ids_any_of: Optional[Iterable[str]] = None,
        sample_owner_ids_none_of: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> Response[ContainersPaginatedList]:
        response = list_containers.sync_detailed(
            client=self.client,
            sort=none_as_unset(sort),
            schema_id=none_as_unset(schema_id),
            modified_at=none_as_unset(modified_at),
            created_at=none_as_unset(created_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            ancestor_storage_id=none_as_unset(ancestor_storage_id),
            storage_contents_id=none_as_unset(storage_contents_id),
            storage_contents_ids=none_as_unset(optional_array_query_param(storage_contents_ids)),
            checkout_status=none_as_unset(checkout_status),
            ids=none_as_unset(optional_array_query_param(ids)),
            barcodes=none_as_unset(optional_array_query_param(barcodes)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            checkout_assignee_idsany_of=none_as_unset(
                optional_array_query_param(checkout_assignee_ids_any_of)
            ),
            restricted_sample_party_idsall_of=none_as_unset(
                optional_array_query_param(restricted_sample_party_ids_all_of)
            ),
            restricted_sample_party_idsany_of=none_as_unset(
                optional_array_query_param(restricted_sample_party_ids_any_of)
            ),
            restricted_sample_party_idsnone_of=none_as_unset(
                optional_array_query_param(restricted_sample_party_ids_none_of)
            ),
            restriction_status=none_as_unset(restriction_status),
            sample_owner_idsall_of=none_as_unset(optional_array_query_param(sample_owner_ids_all_of)),
            sample_owner_idsany_of=none_as_unset(optional_array_query_param(sample_owner_ids_any_of)),
            sample_owner_idsnone_of=none_as_unset(optional_array_query_param(sample_owner_ids_none_of)),
            archive_reason=none_as_unset(archive_reason),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        sort: Optional[Union[str, ListContainersSort]] = None,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ancestor_storage_id: Optional[str] = None,
        storage_contents_id: Optional[str] = None,
        storage_contents_ids: Optional[List[str]] = None,
        checkout_status: Optional[ListContainersCheckoutStatus] = None,
        ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        checkout_assignee_ids_any_of: Optional[Iterable[str]] = None,
        restricted_sample_party_ids_all_of: Optional[Iterable[str]] = None,
        restricted_sample_party_ids_any_of: Optional[Iterable[str]] = None,
        restricted_sample_party_ids_none_of: Optional[Iterable[str]] = None,
        restriction_status: Optional[SampleRestrictionStatus] = None,
        sample_owner_ids_all_of: Optional[Iterable[str]] = None,
        sample_owner_ids_any_of: Optional[Iterable[str]] = None,
        sample_owner_ids_none_of: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> PageIterator[Container]:
        """
        List containers.

        See https://benchling.com/api/reference#/Containers/listContainers
        """
        check_for_csv_bug_fix("storage_contents_ids", storage_contents_ids)

        def api_call(next_token: NextToken) -> Response[ContainersPaginatedList]:
            return self._containers_page(
                sort=_translate_to_string_enum(ListContainersSort, sort),
                schema_id=schema_id,
                modified_at=modified_at,
                created_at=created_at,
                name=name,
                name_includes=name_includes,
                ancestor_storage_id=ancestor_storage_id,
                storage_contents_id=storage_contents_id,
                storage_contents_ids=storage_contents_ids,
                checkout_status=checkout_status,
                ids=ids,
                barcodes=barcodes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                creator_ids=creator_ids,
                checkout_assignee_ids_any_of=checkout_assignee_ids_any_of,
                restricted_sample_party_ids_all_of=restricted_sample_party_ids_all_of,
                restricted_sample_party_ids_any_of=restricted_sample_party_ids_any_of,
                restricted_sample_party_ids_none_of=restricted_sample_party_ids_none_of,
                restriction_status=restriction_status,
                sample_owner_ids_all_of=sample_owner_ids_all_of,
                sample_owner_ids_any_of=sample_owner_ids_any_of,
                sample_owner_ids_none_of=sample_owner_ids_none_of,
                archive_reason=archive_reason,
                next_token=next_token,
                schema_fields=schema_fields,
                page_size=page_size,
                returning=returning,
            )

        def results_extractor(body: ContainersPaginatedList) -> Optional[List[Container]]:
            return body.containers

        return PageIterator(api_call, results_extractor)

    @api_method
    def bulk_get(
        self,
        *,
        container_ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> Optional[List[Container]]:
        """
        Bulk get containers.

        See https://benchling.com/api/reference#/Containers/bulkGetContainers
        """
        container_id_string = optional_array_query_param(container_ids)
        barcode_string = optional_array_query_param(barcodes)
        returning_string = optional_array_query_param(returning)
        response = bulk_get_containers.sync_detailed(
            client=self.client,
            container_ids=none_as_unset(container_id_string),
            barcodes=none_as_unset(barcode_string),
            returning=none_as_unset(returning_string),
        )
        containers_list = model_from_detailed(response)
        return containers_list.containers

    @api_method
    def bulk_create(
        self, containers: Iterable[ContainerCreate]
    ) -> TaskHelper[BulkCreateContainersAsyncTaskResponse]:
        """
        Bulk create containers.

        See https://benchling.com/api/reference#/Containers/bulkCreateContainers
        """
        body = ContainersBulkCreateRequest(list(containers))
        response = bulk_create_containers.sync_detailed(client=self.client, json_body=body)
        return self._task_helper_from_response(response, BulkCreateContainersAsyncTaskResponse)

    @api_method
    def bulk_update(
        self, containers: Iterable[ContainerBulkUpdateItem]
    ) -> TaskHelper[BulkUpdateContainersAsyncTaskResponse]:
        """
        Bulk update containers.

        See https://benchling.com/api/reference#/Containers/bulkUpdateContainers
        """
        body = ContainersBulkUpdateRequest(list(containers))
        response = bulk_update_containers.sync_detailed(client=self.client, json_body=body)
        return self._task_helper_from_response(response, BulkUpdateContainersAsyncTaskResponse)

    @api_method
    def create(self, container: ContainerCreate) -> Container:
        """
        Create a new container.

        See https://benchling.com/api/reference#/Containers/createContainer
        """
        response = create_container.sync_detailed(client=self.client, json_body=container)
        return model_from_detailed(response)

    @api_method
    def update(self, container_id: str, container: ContainerUpdate) -> Container:
        """
        Update a container.

        See https://benchling.com/api/reference#/Containers/updateContainer
        """
        response = update_container.sync_detailed(
            client=self.client, container_id=container_id, json_body=container
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self, container_ids: Iterable[str], reason: ContainersArchiveReason, should_remove_barcodes: bool
    ) -> ContainersArchivalChange:
        """
        Archive containers.

        See https://benchling.com/api/reference#/Containers/archiveContainers
        """
        archive_request = ContainersArchive(
            container_ids=list(container_ids), reason=reason, should_remove_barcodes=should_remove_barcodes
        )
        response = archive_containers.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, container_ids: Iterable[str]) -> ContainersArchivalChange:
        """
        Unarchive containers.

        See https://benchling.com/api/reference#/Containers/unarchiveContainers
        """
        unarchive_request = ContainersUnarchive(container_ids=list(container_ids))
        response = unarchive_containers.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def print_labels(self, print_request: PrintLabels) -> None:
        """
        Print labels.

        See https://benchling.com/api/reference#/Containers/printLabels
        """
        response = print_labels.sync_detailed(client=self.client, json_body=print_request)
        raise_for_status(response)

    @api_method
    def contents_by_id(self, container_id: str, containable_id: str) -> ContainerContent:
        """
        Get a container content.

        See https://benchling.com/api/reference#/Containers/getContainerContent
        """
        response = get_container_content.sync_detailed(
            client=self.client, container_id=container_id, containable_id=containable_id
        )
        return model_from_detailed(response)

    @api_method
    def list_contents(self, container_id: str) -> List[ContainerContent]:
        """
        List a container's contents.

        See https://benchling.com/api/reference#/Containers/listContainerContents
        """
        response = list_container_contents.sync_detailed(client=self.client, container_id=container_id)
        contents_list = model_from_detailed(response)
        return contents_list.contents

    @api_method
    def update_contents(
        self, container_id: str, containable_id: str, concentration: Measurement
    ) -> ContainerContent:
        """
        Update a container's content.

        See https://benchling.com/api/reference#/Containers/updateContainerContent
        """
        update = ContainerContentUpdate(concentration=concentration)
        response = update_container_content.sync_detailed(
            client=self.client, container_id=container_id, containable_id=containable_id, json_body=update
        )
        return model_from_detailed(response)

    @api_method
    def delete_contents(self, container_id: str, containable_id: str) -> None:
        """
        Delete a container's content.

        See https://benchling.com/api/reference#/Containers/deleteContainerContent
        """
        response = delete_container_content.sync_detailed(
            client=self.client, container_id=container_id, containable_id=containable_id
        )
        raise_for_status(response)

    @api_method
    def reserve(self, reservation: ContainersCheckout) -> None:
        """
        Reserve containers.

        See https://benchling.com/api/reference#/Containers/reserveContainers
        """
        response = reserve_containers.sync_detailed(client=self.client, json_body=reservation)
        raise_for_status(response)

    @api_method
    def checkout(self, checkout: ContainersCheckout) -> None:
        """
        Check out containers.

        See https://benchling.com/api/reference#/Containers/checkoutContainers
        """
        response = checkout_containers.sync_detailed(client=self.client, json_body=checkout)
        raise_for_status(response)

    @api_method
    def checkin(self, checkin: ContainersCheckin) -> None:
        """
        Check in containers.

        See https://benchling.com/api/reference#/Containers/checkinContainers
        """
        response = checkin_containers.sync_detailed(client=self.client, json_body=checkin)
        raise_for_status(response)

    @api_method
    def transfer_into_container(
        self, destination_container_id: str, transfer_request: ContainerTransfer
    ) -> None:
        """
        Transfer into container.

        See https://benchling.com/api/reference#/Containers/transferIntoContainer
        """
        response = transfer_into_container.sync_detailed(
            client=self.client, destination_container_id=destination_container_id, json_body=transfer_request
        )
        raise_for_status(response)

    @api_method
    def transfer_into_containers(
        self, transfer_requests: Iterable[MultipleContainersTransfer]
    ) -> TaskHelper[TransfersAsyncTaskResponse]:
        """
        Transfer into containers.

        See https://benchling.com/api/reference#/Containers/transferIntoContainers
        """
        multiple_requests = MultipleContainersTransfersList(transfers=list(transfer_requests))
        response = transfer_into_containers.sync_detailed(client=self.client, json_body=multiple_requests)
        return self._task_helper_from_response(response, TransfersAsyncTaskResponse)



from typing import List, Optional

from benchling_api_client.v2.beta.api.analyses import get_analysis, update_analysis
from benchling_api_client.v2.beta.models.analysis import Analysis
from benchling_api_client.v2.beta.models.analysis_update import AnalysisUpdate

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaAnalysisService(BaseService):
    """
    V2-Beta Analyses.

    Analyses allow experimental data to be viewed, analyzed, and visualized.

    https://benchling.com/api/v2-beta/reference#/Analyses
    """

    @api_method
    def get_by_id(self, analysis_id: str, returning: Optional[List[str]] = None) -> Analysis:
        """
        Get an analysis.

        See https://benchling.com/api/v2-beta/reference#/Analyses/getAnalysis
        """
        response = get_analysis.sync_detailed(
            client=self.client,
            analysis_id=analysis_id,
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        return model_from_detailed(response)

    @api_method
    def update(self, analysis_id: str, update: AnalysisUpdate) -> Analysis:
        """
        Update an analysis.

        See https://benchling.com/api/v2-beta/reference#/Analyses/updateAnalysis
        """
        response = update_analysis.sync_detailed(
            client=self.client, analysis_id=analysis_id, json_body=update
        )
        return model_from_detailed(response)



class UnsupportedConfigItemError(Exception):
    """
    Unsupported config item error.

    The manifest and configuration specified an item which the SDK is incapable of handling yet.
    """

    pass


class MissingRequiredConfigItemError(Exception):
    """
    Missing required config item error.

    A config item was invoked as required, but missing in the configuration item store.
    """

    pass


class InaccessibleConfigItemError(Exception):
    """
    Inaccessible config item error.

    A config item linked resource was inaccessible (the caller does not have permissions).
    """

    pass


class ConfigItemLinkedResourceError(Exception):
    """
    Config item linked resource error.

    A config item is not a type that has an associated linked resource.
    """

    pass



from __future__ import annotations

from functools import cached_property
from typing import Optional, TYPE_CHECKING

from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.retry_helpers import RetryStrategy
from benchling_sdk.services.v2.base_service import BaseService

if TYPE_CHECKING:
    from benchling_sdk.services.v2.stable.aa_sequence_service import AaSequenceService
    from benchling_sdk.services.v2.stable.api_service import ApiService
    from benchling_sdk.services.v2.stable.app_service import AppService
    from benchling_sdk.services.v2.stable.assay_result_service import AssayResultService
    from benchling_sdk.services.v2.stable.assay_run_service import AssayRunService
    from benchling_sdk.services.v2.stable.audit_service import AuditService
    from benchling_sdk.services.v2.stable.blob_service import BlobService
    from benchling_sdk.services.v2.stable.box_service import BoxService
    from benchling_sdk.services.v2.stable.codon_usage_table_service import CodonUsageTableService
    from benchling_sdk.services.v2.stable.connect_service import ConnectService
    from benchling_sdk.services.v2.stable.container_service import ContainerService
    from benchling_sdk.services.v2.stable.custom_entity_service import CustomEntityService
    from benchling_sdk.services.v2.stable.custom_notation_service import CustomNotationService
    from benchling_sdk.services.v2.stable.data_frame_service import DataFrameService
    from benchling_sdk.services.v2.stable.dataset_service import DatasetService
    from benchling_sdk.services.v2.stable.dna_alignments_service import DnaAlignmentsService
    from benchling_sdk.services.v2.stable.dna_oligo_service import DnaOligoService
    from benchling_sdk.services.v2.stable.dna_sequence_service import DnaSequenceService
    from benchling_sdk.services.v2.stable.dropdown_service import DropdownService
    from benchling_sdk.services.v2.stable.entity_service import EntityService
    from benchling_sdk.services.v2.stable.entry_service import EntryService
    from benchling_sdk.services.v2.stable.enzyme_service import EnzymeService
    from benchling_sdk.services.v2.stable.event_service import EventService
    from benchling_sdk.services.v2.stable.export_service import ExportService
    from benchling_sdk.services.v2.stable.feature_library_service import FeatureLibraryService
    from benchling_sdk.services.v2.stable.file_service import FileService
    from benchling_sdk.services.v2.stable.folder_service import FolderService
    from benchling_sdk.services.v2.stable.instrument_query_service import InstrumentQueryService
    from benchling_sdk.services.v2.stable.inventory_service import InventoryService
    from benchling_sdk.services.v2.stable.lab_automation_service import LabAutomationService
    from benchling_sdk.services.v2.stable.label_template_service import LabelTemplateService
    from benchling_sdk.services.v2.stable.legacy_request_service import LegacyRequestService
    from benchling_sdk.services.v2.stable.location_service import LocationService
    from benchling_sdk.services.v2.stable.mixture_service import MixtureService
    from benchling_sdk.services.v2.stable.molecule_service import MoleculeService
    from benchling_sdk.services.v2.stable.monomer_service import MonomerService
    from benchling_sdk.services.v2.stable.nucleotide_alignments_service import NucleotideAlignmentsService
    from benchling_sdk.services.v2.stable.oligo_service import OligoService
    from benchling_sdk.services.v2.stable.organization_service import OrganizationService
    from benchling_sdk.services.v2.stable.plate_service import PlateService
    from benchling_sdk.services.v2.stable.printer_service import PrinterService
    from benchling_sdk.services.v2.stable.project_service import ProjectService
    from benchling_sdk.services.v2.stable.registry_service import RegistryService
    from benchling_sdk.services.v2.stable.rna_oligo_service import RnaOligoService
    from benchling_sdk.services.v2.stable.rna_sequence_service import RnaSequenceService
    from benchling_sdk.services.v2.stable.schema_service import SchemaService
    from benchling_sdk.services.v2.stable.task_service import TaskService
    from benchling_sdk.services.v2.stable.team_service import TeamService
    from benchling_sdk.services.v2.stable.test_order_service import TestOrderService
    from benchling_sdk.services.v2.stable.user_service import UserService
    from benchling_sdk.services.v2.stable.warehouse_service import WarehouseService
    from benchling_sdk.services.v2.stable.workflow_flowchart_config_version_service import (
        WorkflowFlowchartConfigVersionService,
    )
    from benchling_sdk.services.v2.stable.workflow_flowchart_service import WorkflowFlowchartService
    from benchling_sdk.services.v2.stable.workflow_output_service import WorkflowOutputService
    from benchling_sdk.services.v2.stable.workflow_task_group_service import WorkflowTaskGroupService
    from benchling_sdk.services.v2.stable.workflow_task_service import WorkflowTaskService


class V2StableService(BaseService):
    """
    V2 Stable.

    Namespace containing support for the V2 stable endpoints of the Benchling API.
    """

    def __init__(self, client: Client, retry_strategy: Optional[RetryStrategy] = None):
        """
        Initialize a service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        super().__init__(client, retry_strategy)

    @cached_property
    def aa_sequences(self) -> AaSequenceService:
        """
        AA Sequences.

        AA Sequences are the working units of cells that make everything run (they help make structures, catalyze
        reactions and allow for signaling - a kind of internal cell communication). On Benchling, these are comprised
        of a string of amino acids and collections of other attributes, such as annotations.

        See https://benchling.com/api/reference#/AA%20Sequences
        """
        from .stable.aa_sequence_service import AaSequenceService

        return self._create_service(AaSequenceService)

    @cached_property
    def api(self) -> ApiService:
        """
        Make custom API calls with the underlying BenchlingApiClient.

        A common use case for this is making calls to API endpoints which may not yet be supported in the current SDK
        release. It's capable of making more "generic" calls utilizing our authorization scheme, as well as supporting
        some simple serialization and deserialization for custom models.
        """
        from .stable.api_service import ApiService

        return self._create_service(ApiService)

    @cached_property
    def apps(self) -> AppService:
        """
        Apps.

        Apps provide a framework for you to customize your teams’ experiences on
        Benchling with custom applications.

        See https://benchling.com/api/reference#/Apps
        and https://docs.benchling.com/docs/getting-started-benchling-apps
        """  # noqa:RUF002  # Ruff gets confused by a trailing apostrophe with a plural noun
        from .stable.app_service import AppService

        return self._create_service(AppService)

    @cached_property
    def assay_results(self) -> AssayResultService:
        """
        Assay Results.

        Results represent the output of assays that have been performed. You can customize the schemas of results to
        fit your needs. Results can link to runs, batches, and other types.

        See https://benchling.com/api/reference#/Assay%20Results
        """
        from .stable.assay_result_service import AssayResultService

        return self._create_service(AssayResultService)

    @cached_property
    def assay_runs(self) -> AssayRunService:
        """
        Assay Runs.

        Runs capture the details / parameters of a run that was performed. Results are usually nested under a run.

        See https://benchling.com/api/reference#/Assay%20Runs
        """
        from .stable.assay_run_service import AssayRunService

        return self._create_service(AssayRunService)

    @cached_property
    def audit(self) -> AuditService:
        """
        Audits.

        Export audit log data for Benchling objects.

        https://benchling.com/api/reference#/Audit
        """
        from .stable.audit_service import AuditService

        return self._create_service(AuditService)

    @cached_property
    def blobs(self) -> BlobService:
        """
        Blobs.

        Blobs are opaque files that can be linked to other items in Benchling, like assay runs or results. For example,
        you can upload a blob, then upload an assay result that links to that blob by ID. The blob will then appear as
        part of the assay result in the Benchling web UI.

        See https://benchling.com/api/reference#/Blobs
        """
        from .stable.blob_service import BlobService

        return self._create_service(BlobService)

    @cached_property
    def boxes(self) -> BoxService:
        """
        Boxes.

        Boxes are a structured storage type, consisting of a grid of positions that can each hold one container. Unlike
        locations, there are a maximum number of containers that a box can hold (one per position).

        Boxes are all associated with schemas, which define the type of the box (e.g. "10x10 Cryo Box") along with the
        fields that are tracked and the dimensions of the box.

        Like all storage, every Box has a barcode that is unique across the registry.

        See https://benchling.com/api/reference#/Boxes
        """
        from .stable.box_service import BoxService

        return self._create_service(BoxService)

    @cached_property
    def codon_usage_tables(self) -> CodonUsageTableService:
        """
        Codon Usage Tables.

        Benchling curates codon usage data for a variety of organisms to support operations such as Codon
        Optimization and Back Translation.

        See https://benchling.com/api/reference#/Codon%20Usage%20Tables
        """
        from .stable.codon_usage_table_service import CodonUsageTableService

        return self._create_service(CodonUsageTableService)

    @cached_property
    def connect(self) -> ConnectService:
        """
        Connect.

        Connect endpoints support Benchling Connect actions, like instrument data conversion.

        See https://benchling.com/api/reference#/Connect
        """
        from .stable.connect_service import ConnectService

        return self._create_service(ConnectService)

    @cached_property
    def containers(self) -> ContainerService:
        """
        Containers.

        Containers are the backbone of sample management in Benchling. They represent physical containers, such as
        tubes or wells, that hold quantities of biological samples (represented by the batches inside the container).
        The container itself tracks its total volume, and the concentration of every batch inside of it.

        Containers are all associated with schemas, which define the type of the container (e.g. "Tube") along with the
        fields that are tracked.

        Like all storage, every container has a barcode that is unique across the registry.

        See https://benchling.com/api/reference#/Containers
        """
        from .stable.container_service import ContainerService

        return self._create_service(ContainerService)

    @cached_property
    def custom_entities(self) -> CustomEntityService:
        """
        Custom Entities.

        Benchling supports custom entities for biological entities that are neither sequences or proteins. Custom
        entities must have an entity schema set and can have both schema fields and custom fields.

        See https://benchling.com/api/reference#/Custom%20Entities
        """
        from .stable.custom_entity_service import CustomEntityService

        return self._create_service(CustomEntityService)

    @cached_property
    def custom_notations(self) -> CustomNotationService:
        """
        Custom Notations.

        Benchling allows users to configure their own fully-custom string representation formats for import/export
        of nucleotide sequences (including chemical modifications).

        See https://benchling.com/api/reference#/Custom%20Notations
        """
        from .stable.custom_notation_service import CustomNotationService

        return self._create_service(CustomNotationService)

    @cached_property
    def data_frames(self) -> DataFrameService:
        """
        DataFrames.

        DataFrames are Benchling objects that represent tabular data with typed columns and rows of data.

        See https://benchling.com/api/v2/reference#/Data%20Frames
        """
        from .stable.data_frame_service import DataFrameService

        return self._create_service(DataFrameService)

    @cached_property
    def datasets(self) -> DatasetService:
        """
        Datasets.

        Datasets are Benchling objects that represent tabular data with typed columns and rows of data. Unlike
        Data Frames, Datasets are located in folders and can be searched in the UI.

        See https://benchling.com/api/v2/reference#/Datasets
        """
        from .stable.dataset_service import DatasetService

        return self._create_service(DatasetService)

    @cached_property
    def dna_alignments(self) -> DnaAlignmentsService:
        """
        DNA Alignments.

        A DNA alignment is a Benchling object representing an alignment of multiple DNA sequences.

        See https://benchling.com/api/reference#/DNA%20Alignments
        """
        from .stable.dna_alignments_service import DnaAlignmentsService

        return self._create_service(DnaAlignmentsService)

    @cached_property
    def dna_oligos(self) -> DnaOligoService:
        """
        DNA Oligos.

        DNA Oligos are short linear DNA sequences that can be attached as primers to full DNA sequences. Just like other
        entities, they support schemas, tags, and aliases.

        See https://benchling.com/api/reference#/DNA%20Oligos
        """
        from .stable.dna_oligo_service import DnaOligoService

        return self._create_service(DnaOligoService)

    @cached_property
    def dna_sequences(self) -> DnaSequenceService:
        """
        DNA Sequences.

        DNA sequences are the bread and butter of the Benchling Molecular Biology suite. On Benchling, these are
        comprised of a string of nucleotides and collections of other attributes, such as annotations and primers.

        See https://benchling.com/api/reference#/DNA%20Sequences
        """
        from .stable.dna_sequence_service import DnaSequenceService

        return self._create_service(DnaSequenceService)

    @cached_property
    def rna_sequences(self) -> RnaSequenceService:
        """
        RNA Sequences.

        Chains of linear, single stranded RNA that support most capabilities and attributes of DNA Sequences.

        See https://benchling.com/api/reference#/RNA%20Sequences
        """
        from .stable.rna_sequence_service import RnaSequenceService

        return self._create_service(RnaSequenceService)

    @cached_property
    def dropdowns(self) -> DropdownService:
        """
        Dropdowns.

        Dropdowns are registry-wide enums. Use dropdowns to standardize on spelling and naming conventions, especially
        for important metadata like resistance markers.

        See https://benchling.com/api/reference#/Dropdowns
        """
        from .stable.dropdown_service import DropdownService

        return self._create_service(DropdownService)

    @cached_property
    def entities(self) -> EntityService:
        """
        Entities.

        Entities include DNA and AA sequences, oligos, molecules, custom entities, and
        other biological objects in Benchling. Entities support schemas, tags, and aliases,
        and can be registered.

        See https://benchling.com/api/reference#/Entities
        """
        from .stable.entity_service import EntityService

        return self._create_service(EntityService)

    @cached_property
    def entries(self) -> EntryService:
        """
        Entries.

        Entries are rich text documents that allow you to capture all of your experimental data in one place.

        See https://benchling.com/api/reference#/Entries
        """
        from .stable.entry_service import EntryService

        return self._create_service(EntryService)

    @cached_property
    def enzymes(self) -> EnzymeService:
        """
        Enzymes.

        Restriction enzymes are curated by Benchling for operations such as Digests and Codon Optimization.

        See https://benchling.com/api/reference#/Enzymes
        """
        from .stable.enzyme_service import EnzymeService

        return self._create_service(EnzymeService)

    @cached_property
    def events(self) -> EventService:
        """
        Events.

        The Events system allows external services to subscribe to events that are triggered in Benchling (e.g. plasmid
        registration, request submission, etc).

        See https://benchling.com/api/reference#/Events
        """
        from .stable.event_service import EventService

        return self._create_service(EventService)

    @cached_property
    def exports(self) -> ExportService:
        """
        Exports.

        Export a Notebook Entry.

        See https://benchling.com/api/reference#/Exports
        """
        from .stable.export_service import ExportService

        return self._create_service(ExportService)

    @cached_property
    def feature_libraries(self) -> FeatureLibraryService:
        """
        Feature Libraries.

        Feature Libraries are collections of shared canonical patterns that can be used to generate
        annotations on matching regions of DNA Sequences or AA Sequences.

        See https://benchling.com/api/reference#/Feature%20Libraries
        """
        from .stable.feature_library_service import FeatureLibraryService

        return self._create_service(FeatureLibraryService)

    @cached_property
    def files(self) -> FileService:
        """
        Files.

        Files are Benchling objects that represent files and their metadata. Compared to Blobs, which are used
        by most Benchling products for attachments, Files are primarily used in the Analysis and Connect
        product.

        See https://benchling.com/api/v2/reference#/Files
        """
        from .stable.file_service import FileService

        return self._create_service(FileService)

    @cached_property
    def folders(self) -> FolderService:
        """
        Folders.

        Manage folder objects.

        See https://benchling.com/api/reference#/Folders
        """
        from .stable.folder_service import FolderService

        return self._create_service(FolderService)

    @cached_property
    def instrument_queries(self) -> InstrumentQueryService:
        """
        Instrument Queries.

        Instrument Queries are used to query the instrument service.

        See https://benchling.com/api/reference#/Instrument%20Queries
        """
        from .stable.instrument_query_service import InstrumentQueryService

        return self._create_service(InstrumentQueryService)

    @cached_property
    def inventory(self) -> InventoryService:
        """
        Inventory.

        Manage inventory wide objects.

        See https://benchling.com/api/reference#/Inventory
        """
        from .stable.inventory_service import InventoryService

        return self._create_service(InventoryService)

    @cached_property
    def lab_automation(self) -> LabAutomationService:
        """
        Lab Automation.

        Lab Automation endpoints support integration with lab instruments, and liquid handlers to create samples or
        results, and capture transfers between containers at scale.

        See https://benchling.com/api/reference#/Lab%20Automation
        """
        from .stable.lab_automation_service import LabAutomationService

        return self._create_service(LabAutomationService)

    @cached_property
    def label_templates(self) -> LabelTemplateService:
        """
        Label Templates.

        List label templates.

        See https://benchling.com/api/reference#/Label%20Templates
        """
        from .stable.label_template_service import LabelTemplateService

        return self._create_service(LabelTemplateService)

    @cached_property
    def legacy_requests(self) -> LegacyRequestService:
        """
        Legacy Requests.

        Legacy Requests allow scientists and teams to collaborate around experimental assays and workflows.

        See https://benchling.com/api/reference#/Legacy%20Requests
        """
        from .stable.legacy_request_service import LegacyRequestService

        return self._create_service(LegacyRequestService)

    @cached_property
    def locations(self) -> LocationService:
        """
        Locations.

        Manage locations objects. Like all storage, every Location has a barcode that is unique across the registry.

        See https://benchling.com/api/reference#/Locations
        """
        from .stable.location_service import LocationService

        return self._create_service(LocationService)

    @cached_property
    def mixtures(self) -> MixtureService:
        """
        Mixtures.

        Mixtures are solutions comprised of multiple ingredients where the exact quantities of each ingredient are
        important to track. Each ingredient is uniquely identified by its component entity.

        See https://benchling.com/api/reference#/Mixtures
        """
        from .stable.mixture_service import MixtureService

        return self._create_service(MixtureService)

    @cached_property
    def monomers(self) -> MonomerService:
        """
        Monomers.

        Monomers are chemical building blocks with specified structures used to compose modified
        nucleotides. Note that monomer write endpoints require tenant admin permissions.

        See https://benchling.com/api/reference#/Monomers
        """
        from .stable.monomer_service import MonomerService

        return self._create_service(MonomerService)

    @cached_property
    def molecules(self) -> MoleculeService:
        """
        Molecules.

        Molecules are groups of atoms held together by bonds, representing entities smaller than DNA
        Sequences and AA Sequences. Just like other entities, they support schemas, tags, and aliases.

        See https://benchling.com/api/reference#/Molecules
        """
        from .stable.molecule_service import MoleculeService

        return self._create_service(MoleculeService)

    @cached_property
    def nucleotide_alignments(self) -> NucleotideAlignmentsService:
        """
        Nucleotide Alignments.

        A Nucleotide Alignment is a Benchling object representing an alignment of multiple DNA and/or RNA sequences.

        See https://benchling.com/api/reference#/Nucleotide%20Alignments
        """
        from .stable.nucleotide_alignments_service import NucleotideAlignmentsService

        return self._create_service(NucleotideAlignmentsService)

    @cached_property
    def oligos(self) -> OligoService:
        """
        Oligos.

        Oligos are short linear DNA sequences that can be attached as primers to full DNA sequences. Just like other
        entities, they support schemas, tags, and aliases.

        Please migrate to the corresponding DNA Oligos endpoints so that we can support RNA Oligos.

        See https://benchling.com/api/reference#/Oligos
        """
        from .stable.oligo_service import OligoService

        return self._create_service(OligoService)

    @cached_property
    def organizations(self) -> OrganizationService:
        """
        Organizations.

        View organization objects.

        See https://benchling.com/api/reference#/Organizations
        """
        from .stable.organization_service import OrganizationService

        return self._create_service(OrganizationService)

    @cached_property
    def plates(self) -> PlateService:
        """
        Plates.

        Plates are a structured storage type, grids of wells that each function like containers. Plates come in two
        types: a traditional "fixed" type, where the wells cannot move, and a "matrix" type. A matrix plate has similar
        functionality to a box, where the containers inside can be moved around and removed altogether.

        Plates are all associated with schemas, which define the type of the plate (e.g. "96 Well Plate") along with
        the fields that are tracked, the dimensions of the plate, and whether or not the plate is a matrix plate or a
        traditional well plate.

        Like all storage, every Plate has a barcode that is unique across the registry.

        See https://benchling.com/api/reference#/Plates
        """
        from .stable.plate_service import PlateService

        return self._create_service(PlateService)

    @cached_property
    def printers(self) -> PrinterService:
        """
        Printers.

        List printers.

        See https://benchling.com/api/reference#/Printers
        """
        from .stable.printer_service import PrinterService

        return self._create_service(PrinterService)

    @cached_property
    def projects(self) -> ProjectService:
        """
        Projects.

        Manage project objects.

        See https://benchling.com/api/reference#/Projects
        """
        from .stable.project_service import ProjectService

        return self._create_service(ProjectService)

    @cached_property
    def registry(self) -> RegistryService:
        """
        Registry.

        Manage registry objects.

        See https://benchling.com/api/reference#/Registry
        """
        from .stable.registry_service import RegistryService

        return self._create_service(RegistryService)

    @cached_property
    def rna_oligos(self) -> RnaOligoService:
        """
        RNA Oligos.

        RNA Oligos are short linear RNA sequences that can be attached as primers to full DNA sequences. Just like other
        entities, they support schemas, tags, and aliases.

        See https://benchling.com/api/reference#/RNA%20Oligos
        """
        from .stable.rna_oligo_service import RnaOligoService

        return self._create_service(RnaOligoService)

    @cached_property
    def schemas(self) -> SchemaService:
        """
        Schemas.

        Schemas represent custom configuration of objects in Benchling. See https://docs.benchling.com/docs/schemas in
        our documentation on how Schemas impact our developers

        See https://benchling.com/api/reference#/Schemas
        """
        from .stable.schema_service import SchemaService

        return self._create_service(SchemaService)

    @cached_property
    def tasks(self) -> TaskService:
        """
        Tasks.

        Endpoints that perform expensive computations launch long-running tasks. These endpoints return the task ID (a
        UUID) in the response body.

        After launching a task, periodically invoke the Get a task endpoint with the task UUID (e.g., every 10
        seconds), until the status is no longer RUNNING.

        You can access a task for up to 30 minutes after its completion, after which its data will no longer be
        available.

        See https://benchling.com/api/reference#/Tasks
        """
        from .stable.task_service import TaskService

        return self._create_service(TaskService)

    @cached_property
    def teams(self) -> TeamService:
        """
        Teams.

        View team objects.

        See https://benchling.com/api/reference#/Teams
        """
        from .stable.team_service import TeamService

        return self._create_service(TeamService)

    @cached_property
    def test_orders(self) -> TestOrderService:
        """
        Test Orders.

        Test orders enable users to order tests for specific sample/container combinations that will be fulfilled in assays.

        See https://benchling.com/api/reference?availability=la#/Test%20Orders/
        """
        from .stable.test_order_service import TestOrderService

        return self._create_service(TestOrderService)

    @cached_property
    def users(self) -> UserService:
        """
        Benchling users.

        See https://benchling.com/api/reference#/Users
        """
        from .stable.user_service import UserService

        return self._create_service(UserService)

    @cached_property
    def warehouse(self) -> WarehouseService:
        """
        Warehouse.

        Manage warehouse credentials.

        See https://benchling.com/api/reference#/Warehouse
        """
        from .stable.warehouse_service import WarehouseService

        return self._create_service(WarehouseService)

    @cached_property
    def workflow_flowcharts(self) -> WorkflowFlowchartService:
        """
        Workflow Flowcharts.

        Workflow flowcharts represent the nodes and edges that a flowchart is comprised of.

        See https://benchling.com/api/reference#/Workflow%20Flowcharts
        """
        from .stable.workflow_flowchart_service import WorkflowFlowchartService

        return self._create_service(WorkflowFlowchartService)

    @cached_property
    def workflow_flowchart_config_versions(self) -> WorkflowFlowchartConfigVersionService:
        """
        Workflow Flowchart Config Versions.

        Workflow flowchart config versions are versioned graphs of flowchart configurations.

        See https://benchling.com/api/reference#/Workflow%20Flowchart%20Config%20Versions
        """
        from .stable.workflow_flowchart_config_version_service import WorkflowFlowchartConfigVersionService

        return self._create_service(WorkflowFlowchartConfigVersionService)

    @cached_property
    def workflow_outputs(self) -> WorkflowOutputService:
        """
        Workflow Outputs.

        Workflow outputs are outputs of a workflow task.

        See https://benchling.com/api/reference#/Workflow%20Outputs
        """
        from .stable.workflow_output_service import WorkflowOutputService

        return self._create_service(WorkflowOutputService)

    @cached_property
    def workflow_task_groups(self) -> WorkflowTaskGroupService:
        """
        Workflow Tasks Groups.

        Workflow task groups are groups of workflow tasks of the same schema.

        See https://benchling.com/api/reference#/Workflow%20Task%20Groups
        """
        from .stable.workflow_task_group_service import WorkflowTaskGroupService

        return self._create_service(WorkflowTaskGroupService)

    @cached_property
    def workflow_tasks(self) -> WorkflowTaskService:
        """
        Workflow Tasks.

        Workflow tasks encapsulate a single unit of work.

        See https://benchling.com/api/reference#/Workflow%20Tasks
        """
        from .stable.workflow_task_service import WorkflowTaskService

        return self._create_service(WorkflowTaskService)



from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.custom_entities import (
    archive_custom_entities,
    bulk_create_custom_entities,
    bulk_get_custom_entities,
    bulk_update_custom_entities,
    bulk_upsert_custom_entities,
    create_custom_entity,
    get_custom_entity,
    list_custom_entities,
    unarchive_custom_entities,
    update_custom_entity,
    upsert_custom_entity,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import (
    BulkCreateCustomEntitiesAsyncTaskResponse,
    BulkUpdateCustomEntitiesAsyncTaskResponse,
    CustomEntitiesArchivalChange,
    CustomEntitiesArchive,
    CustomEntitiesBulkCreateRequest,
    CustomEntitiesBulkUpdateRequest,
    CustomEntitiesBulkUpsertRequest,
    CustomEntitiesPaginatedList,
    CustomEntitiesUnarchive,
    CustomEntity,
    CustomEntityBulkCreate,
    CustomEntityBulkUpdate,
    CustomEntityCreate,
    CustomEntityUpdate,
    CustomEntityUpsertRequest,
    EntityArchiveReason,
    ListCustomEntitiesSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class CustomEntityService(BaseService):
    """
    Service for managing custom entities in Benchling.

    Custom entities are flexible biological entities that allow you to model 
    lab-specific objects like cell lines, reagents, samples, or any other 
    biological entity that doesn't fit standard sequence types.

    All custom entities must have an entity schema that defines their structure, 
    validation rules, and field types. They support both schema fields (defined 
    by the schema) and custom fields (user-defined metadata).

    See https://benchling.com/api/reference#/Custom%20Entities
    """

    @api_method
    def get_by_id(self, entity_id: str, returning: Optional[Iterable[str]] = None) -> CustomEntity:
        """
        Get a custom entity.

        Args:
            entity_id (str): The ID of the custom entity to get.
            returning (Optional[Iterable[str]]): The fields to return.
                See https://benchling.com/api/reference#/Custom%20Entities/getCustomEntity
                for available fields. If not specified, returns all fields. Examples:
                - ["id", "name", "fields"] for basic info only
                - ["customEntities.fields.Status"] for specific field

        Returns:
            CustomEntity: The custom entity.

        Example:
            Get a custom entity with all fields:

            .. code-block:: python

                entity = custom_entities.get_by_id("bfi_abc123")
                print(f"Entity: {entity.name} (Schema: {entity.schema_id})")
                
                # Get only specific fields for performance
                entity = custom_entities.get_by_id(
                    "bfi_abc123", 
                    returning=["id", "name", "fields.Vendor"]
                )

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/getCustomEntity

        """
        returning_string = optional_array_query_param(returning)
        response = get_custom_entity.sync_detailed(
            client=self.client, custom_entity_id=entity_id, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)

    @api_method
    def _custom_entities_page(
        self,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        sort: Optional[ListCustomEntitiesSort] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> Response[CustomEntitiesPaginatedList]:
        response = list_custom_entities.sync_detailed(
            client=self.client,
            schema_id=none_as_unset(schema_id),
            modified_at=none_as_unset(modified_at),
            created_at=none_as_unset(created_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            project_id=none_as_unset(project_id),
            registry_id=none_as_unset(registry_id),
            archive_reason=none_as_unset(archive_reason),
            mentions=none_as_unset(optional_array_query_param(mentions)),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            entity_registry_idsany_of=none_as_unset(optional_array_query_param(entity_registry_ids_any_of)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            author_idsany_of=none_as_unset(optional_array_query_param(author_idsany_of)),
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        sort: Optional[Union[str, ListCustomEntitiesSort]] = None,
        page_size: Optional[int] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> PageIterator[CustomEntity]:
        """
        List custom entities with flexible filtering and pagination.

        This method supports comprehensive filtering options to find specific entities.
        For performance on large datasets, always specify schema_id when possible.

        Args:
            schema_id (Optional[str]): Filter to entities of this schema type. Recommended
                for performance on tenants with many entities.
            modified_at (Optional[str]): Filter to entities modified after this date/time.
                Format: ISO 8601 (e.g., "2024-01-15" or "2024-01-15T10:30:00Z").
            created_at (Optional[str]): Filter to entities created after this date/time.
                Same format as modified_at.
            name (Optional[str]): Filter to entities with this exact name (case-sensitive).
            name_includes (Optional[str]): Filter to entities whose name contains this substring
                (case-insensitive).
            folder_id (Optional[str]): Filter to entities in this folder.
            mentioned_in (Optional[List[str]]): Filter to entities mentioned in these entries.
            project_id (Optional[str]): Filter to entities in this project.
            registry_id (Optional[str]): Filter to entities in this registry.
            archive_reason (Optional[str]): Filter to entities archived for this reason.
            mentions (Optional[List[str]]): Filter to entities that mention these items.
            ids (Optional[Iterable[str]]): Filter to entities with these specific IDs.
            entity_registry_ids_any_of (Optional[Iterable[str]]): Filter to entities with
                any of these registry IDs.
            names_any_of (Optional[Iterable[str]]): Filter to entities with any of these
                names (case-insensitive).
            names_any_of_case_sensitive (Optional[Iterable[str]]): Filter to entities with
                any of these names (case-sensitive).
            creator_ids (Optional[Iterable[str]]): Filter to entities created by these users.
            schema_fields (Optional[Dict[str, Any]]): Filter by schema field values.
                Example: {"Status": "Active", "Priority": "High"}.
            sort (Optional[Union[str, ListCustomEntitiesSort]]): Sort order.
                Examples: "name:asc", "modifiedAt:desc".
            page_size (Optional[int]): Number of results per page (max 100).
            author_idsany_of (Optional[Iterable[str]]): Filter to entities authored by these users.
            returning (Optional[Iterable[str]]): Specify which fields to return for performance.
                Examples: ["id", "name"], ["customEntities.fields.Status"].

        Returns:
            PageIterator[CustomEntity]: An iterator over pages of custom entities.

        Example:
            List entities with filtering and pagination:

            .. code-block:: python

                entities = custom_entities.list(
                    schema_id="ts_cellline_schema",
                    name_includes="HEK",
                    folder_id="lib_abc123",
                    modified_at="2024-01-01",
                    page_size=50,
                    sort="name:asc",
                    returning=["id", "name", "fields.Status"]
                )
                
                for page in entities:
                    for entity in page:
                        print(f"Entity: {entity.name} (ID: {entity.id})")

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/listCustomEntities

        """
        check_for_csv_bug_fix("mentioned_in", mentioned_in)
        check_for_csv_bug_fix("mentions", mentions)

        if returning and "nextToken" not in returning:
            returning = list(returning) + ["nextToken"]

        def api_call(next_token: NextToken) -> Response[CustomEntitiesPaginatedList]:
            return self._custom_entities_page(
                schema_id=schema_id,
                modified_at=modified_at,
                created_at=created_at,
                name=name,
                name_includes=name_includes,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                project_id=project_id,
                registry_id=registry_id,
                archive_reason=archive_reason,
                mentions=mentions,
                ids=ids,
                entity_registry_ids_any_of=entity_registry_ids_any_of,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                creator_ids=creator_ids,
                schema_fields=schema_fields,
                sort=_translate_to_string_enum(ListCustomEntitiesSort, sort),
                page_size=page_size,
                next_token=next_token,
                author_idsany_of=author_idsany_of,
                returning=returning,
            )

        def results_extractor(body: CustomEntitiesPaginatedList) -> Optional[List[CustomEntity]]:
            return body.custom_entities

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, entity: CustomEntityCreate) -> CustomEntity:
        """
        Create a new custom entity.

        Creates a single custom entity with the specified properties. The entity
        must be associated with a valid schema and folder.

        Args:
            entity (CustomEntityCreate): 
                The custom entity to create. Contains the following fields:
                
                **Required Fields:**
                
                * **name** (str): The display name for the entity
                * **schema_id** (str): The ID of the schema that defines this entity type  
                * **folder_id** (str): The ID of the folder where the entity will be stored
                
                **Optional Fields:**
                
                * **fields** (Optional[Fields]): Schema-defined field values created using the 
                  fields() helper. Structure: ``{"field_name": {"value": field_value}}``
                * **custom_fields** (Optional[CustomFields]): User-defined field values created 
                  using custom_fields() helper
                * **registry_id** (Optional[str]): The ID of the registry if the entity should 
                  be registered
                * **entity_registry_id** (Optional[str]): Custom identifier for the entity in 
                  the registry
                * **naming_strategy** (Optional[NamingStrategy]): Strategy for naming registered 
                  entities (e.g., ``NamingStrategy.KEEP_NAMES``)
                * **author_ids** (Optional[List[str]]): List of user IDs to set as authors
                * **custom_notation_ids** (Optional[List[str]]): List of custom notation IDs

        Returns:
            CustomEntity: The newly created custom entity with generated ID and server-populated fields.

        Note:
            For registry entities, provide either:
            
            * ``registry_id`` + ``entity_registry_id`` OR
            * ``registry_id`` + ``naming_strategy``
            
            You cannot specify both ``entity_registry_id`` and ``naming_strategy``.

        Example:
            Create a custom entity with schema and custom fields:

            .. code-block:: python

                from benchling_sdk.models import CustomEntityCreate, NamingStrategy
                from benchling_sdk.helpers.serialization_helpers import fields, custom_fields

                entity = CustomEntityCreate(
                    name="Cell Line HEK293",
                    schema_id="ts_rbQWr8Pf",
                    folder_id="lib_TL5mqoz9",
                    registry_id="src_LbZzJIke", 
                    naming_strategy=NamingStrategy.NEW_IDS,
                    fields=fields({
                        "Vendor": {"value": "Vendor Name"},
                        "Passage Number": {"value": 15}
                    }),
                    custom_fields=custom_fields({
                        "Notes": {"value": "Created via API"}
                    })
                )
                created_entity = custom_entities.create(entity)

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/createCustomEntity

        """
        response = create_custom_entity.sync_detailed(client=self.client, json_body=entity)
        return model_from_detailed(response)

    @api_method
    def update(self, entity_id: str, entity: CustomEntityUpdate) -> CustomEntity:
        """
        Update an existing custom entity.

        Updates the specified fields of a custom entity. Only provided fields will be
        updated; omitted fields remain unchanged.

        Args:
            entity_id (str): 
                The ID of the custom entity to update.
                
            entity (CustomEntityUpdate): 
                The update data containing:
                
                * **fields** (Optional[Fields]): Schema-defined field values to update using 
                  fields() helper. Only specified fields will be updated.
                * **custom_fields** (Optional[CustomFields]): User-defined field values to 
                  update using custom_fields() helper.
                * **author_ids** (Optional[List[str]]): List of user IDs to set as authors.
                * **custom_notation_ids** (Optional[List[str]]): List of custom notation IDs.

        Returns:
            CustomEntity: The updated custom entity with new field values.

        Example:
            Update custom fields on an entity:

            .. code-block:: python

                from benchling_sdk.models import CustomEntityUpdate
                from benchling_sdk.helpers.serialization_helpers import custom_fields

                update = CustomEntityUpdate(
                    custom_fields=custom_fields({
                        "Notes": {"value": "Updated via API"},
                        "Priority": {"value": "High"}
                    })
                )
                updated_entity = custom_entities.update("ent_abc123", update)

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/updateCustomEntity

        """
        response = update_custom_entity.sync_detailed(
            client=self.client, custom_entity_id=entity_id, json_body=entity
        )
        return model_from_detailed(response)

    @api_method
    def archive(self, entity_ids: Iterable[str], reason: EntityArchiveReason) -> CustomEntitiesArchivalChange:
        """
        Archive custom entities.

        Archives multiple custom entities with a specified reason. Archived entities
        are hidden from normal views but can be restored using unarchive().

        Args:
            entity_ids (Iterable[str]): 
                The IDs of custom entities to archive. Can be a list, tuple, or any iterable of string IDs.
                
            reason (EntityArchiveReason): 
                The reason for archiving. Valid values:
                
                * **EntityArchiveReason.MADE_IN_ERROR**: Entity was created by mistake
                * **EntityArchiveReason.RETIRED**: Entity is no longer in use  
                * **EntityArchiveReason.SHIPPED_TO_COLLABORATOR**: Entity sent externally
                * **EntityArchiveReason.OTHER**: Other reason not listed

        Returns:
            CustomEntitiesArchivalChange: Archive operation result containing the list of successfully archived entity IDs 
            and any errors that occurred during archival.

        Example:
            Archive entities with a specific reason:

            .. code-block:: python

                from benchling_sdk.models import EntityArchiveReason

                entity_ids = ["ent_abc123", "ent_def456", "ent_ghi789"]
                result = custom_entities.archive(
                    entity_ids, 
                    reason=EntityArchiveReason.MADE_IN_ERROR
                )
                print(f"Archived {len(result.custom_entity_ids)} entities")

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/archiveCustomEntities

        """
        archive_request = CustomEntitiesArchive(reason=reason, custom_entity_ids=list(entity_ids))
        response = archive_custom_entities.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, entity_ids: Iterable[str]) -> CustomEntitiesArchivalChange:
        """
        Unarchive (restore) previously archived custom entities.

        Restores archived custom entities, making them visible in normal views again.

        Args:
            entity_ids (Iterable[str]): The IDs of archived custom entities to restore.
                Can be a list, tuple, or any iterable of string IDs.

        Returns:
            CustomEntitiesArchivalChange: Unarchive operation result containing the list of successfully unarchived entity IDs 
            and any errors that occurred during unarchival.

        Example:
            Unarchive multiple entities:

            .. code-block:: python

                entity_ids = ["ent_abc123", "ent_def456"] 
                result = custom_entities.unarchive(entity_ids)
                print(f"Unarchived {len(result.custom_entity_ids)} entities")

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/unarchiveCustomEntities

        """
        unarchive_request = CustomEntitiesUnarchive(custom_entity_ids=list(entity_ids))
        response = unarchive_custom_entities.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_get(
        self, entity_ids: Iterable[str], returning: Optional[Iterable[str]] = None
    ) -> Optional[List[CustomEntity]]:
        """
        Retrieve multiple custom entities by ID in a single request.

        Efficiently fetches multiple custom entities by their IDs. More performant
        than calling get_by_id() multiple times.

        Args:
            entity_ids (Iterable[str]): The IDs of custom entities to retrieve.
                Can be a list, tuple, or any iterable of string IDs.
            returning (Optional[Iterable[str]]): Fields to return for performance
                optimization. If not specified, returns all fields. Examples:
                - ["id", "name", "fields"] for basic info only
                - ["customEntities.fields.Status"] for specific fields

        Returns:
            Optional[List[CustomEntity]]: List of retrieved custom entities, or None if no entities found. 
            Order may not match input order.

        Example:
            Get multiple entities with specific fields:

            .. code-block:: python

                entity_ids = ["ent_abc123", "ent_def456", "ent_ghi789"]
                entities = custom_entities.bulk_get(
                    entity_ids,
                    returning=["id", "name", "fields.Status"]
                )
                
                for entity in entities:
                    print(f"Entity: {entity.name} (Status: {entity.fields.get('Status')})")

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/bulkGetCustomEntities

        """
        entity_id_string = ",".join(entity_ids)
        returning_string = optional_array_query_param(returning)
        response = bulk_get_custom_entities.sync_detailed(
            client=self.client,
            custom_entity_ids=entity_id_string,
            returning=none_as_unset(returning_string),
        )
        custom_entities = model_from_detailed(response)
        return custom_entities.custom_entities

    @api_method
    def bulk_create(
        self, entities: Iterable[CustomEntityBulkCreate]
    ) -> TaskHelper[BulkCreateCustomEntitiesAsyncTaskResponse]:
        """
        Create multiple custom entities in a single asynchronous operation.

        Efficiently creates many custom entities at once. Returns a task that runs
        asynchronously - use the TaskHelper to monitor progress and get results.

        Args:
            entities (Iterable[CustomEntityBulkCreate]): 
                The custom entities to create. Each entity should have the same structure as 
                CustomEntityCreate with name, schema_id, folder_id, fields, and optional custom_fields.

        Returns:
            TaskHelper[BulkCreateCustomEntitiesAsyncTaskResponse]: Async task helper for monitoring the bulk creation operation. Use ``wait_for_completion()``
            or check task status to get results.

        Example:
            Create multiple entities and wait for completion:

            .. code-block:: python

                from benchling_sdk.models import CustomEntityBulkCreate
                from benchling_sdk.helpers.serialization_helpers import fields

                entities = [
                    CustomEntityBulkCreate(
                        name=f"Sample {i}",
                        schema_id="ts_sample_schema",
                        folder_id="lib_samples",
                        fields=fields({"Index": {"value": i}})
                    ) for i in range(10)
                ]

                task_helper = custom_entities.bulk_create(entities)
                task = benchling.tasks.wait_for_task(task_helper.task_id)

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/bulkCreateCustomEntities

        """
        body = CustomEntitiesBulkCreateRequest(list(entities))
        response = bulk_create_custom_entities.sync_detailed(client=self.client, json_body=body)
        return self._task_helper_from_response(response, BulkCreateCustomEntitiesAsyncTaskResponse)

    @api_method
    def bulk_update(
        self, entities: Iterable[CustomEntityBulkUpdate]
    ) -> TaskHelper[BulkUpdateCustomEntitiesAsyncTaskResponse]:
        """
        Update multiple custom entities in a single asynchronous operation.

        Efficiently updates many custom entities at once. Returns a task that runs
        asynchronously - use the TaskHelper to monitor progress and get results.

        Args:
            entities (Iterable[CustomEntityBulkUpdate]): The entity updates to apply.
                Each update should contain the entity ID and the fields to update.

        Returns:
            TaskHelper[BulkUpdateCustomEntitiesAsyncTaskResponse]: Async task helper for monitoring the bulk update operation. Use ``wait_for_completion()``
            or check task status to get results.

        Example:
            Update multiple entities with new field values:

            .. code-block:: python

                from benchling_sdk.models import CustomEntityBulkUpdate
                from benchling_sdk.helpers.serialization_helpers import fields

                updates = [
                    CustomEntityBulkUpdate(
                        id="ent_abc123",
                        fields=fields({"Status": {"value": "Complete"}})
                    ),
                    CustomEntityBulkUpdate(
                        id="ent_def456", 
                        fields=fields({"Status": {"value": "In Progress"}})
                    )
                ]

                task_helper = custom_entities.bulk_update(updates)
                task = benchling.tasks.wait_for_task(task_helper.task_id)

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/bulkUpdateCustomEntities

        """
        body = CustomEntitiesBulkUpdateRequest(list(entities))
        response = bulk_update_custom_entities.sync_detailed(client=self.client, json_body=body)
        return self._task_helper_from_response(response, BulkUpdateCustomEntitiesAsyncTaskResponse)

    @api_method
    def upsert(self, entity_registry_id: str, entity: CustomEntityUpsertRequest) -> CustomEntity:
        """
        Create or update a custom entity based on registry ID.

        Creates a new entity if none exists with the given registry ID, or updates
        the existing entity if found. Useful for synchronizing with external systems.

        Args:
            entity_registry_id (str): The registry ID to use for lookup/creation.
                This should be a unique identifier from your external system.
            entity (CustomEntityUpsertRequest): The entity data containing name,
                schema_id, folder_id, registry_id, fields, and optional custom_fields.

        Returns:
            CustomEntity: The created or updated custom entity.

        Example:
            Upsert entity from external system:

            .. code-block:: python

                from benchling_sdk.models import CustomEntityUpsertRequest
                from benchling_sdk.helpers.serialization_helpers import fields

                external_id = "EXT_SAMPLE_001"
                upsert_request = CustomEntityUpsertRequest(
                    name="External Sample 001",
                    schema_id="ts_sample_schema",
                    folder_id="lib_external_samples",
                    registry_id="sample_registry",
                    fields=fields({
                        "External ID": {"value": external_id},
                        "Status": {"value": "Active"}
                    })
                )

                entity = custom_entities.upsert(external_id, upsert_request)

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/upsertCustomEntity

        """
        response = upsert_custom_entity.sync_detailed(
            client=self.client, entity_registry_id=entity_registry_id, json_body=entity
        )
        return model_from_detailed(response)

    @api_method
    def bulk_upsert(
        self, body: CustomEntitiesBulkUpsertRequest
    ) -> TaskHelper[BulkUpdateCustomEntitiesAsyncTaskResponse]:
        """
        Create or update multiple custom entities based on registry IDs.

        Efficiently performs upsert operations on many entities at once. Each entity
        is created if its registry ID doesn't exist, or updated if found.

        Args:
            body (CustomEntitiesBulkUpsertRequest): Bulk upsert request containing
                the list of entities to upsert, each with registry ID and entity data.

        Returns:
            TaskHelper[BulkUpdateCustomEntitiesAsyncTaskResponse]: Async task helper for monitoring the bulk upsert operation. Use ``wait_for_completion()`` 
            to get the final results.

        Example:
            Bulk upsert multiple entities:

            .. code-block:: python

                from benchling_sdk.models import CustomEntitiesBulkUpsertRequest
                
                upsert_request = CustomEntitiesBulkUpsertRequest(
                    custom_entities=[
                        {"registry_id": "REG_001", "name": "Entity 1", "schema_id": "ts_abc123"},
                        {"registry_id": "REG_002", "name": "Entity 2", "schema_id": "ts_abc123"}
                    ]
                )
                
                task_helper = custom_entities.bulk_upsert(upsert_request)
                task = benchling.tasks.wait_for_task(task_helper.task_id)

        See Also:
            https://benchling.com/api/reference#/Custom%20Entities/bulkUpsertCustomEntities

        """
        response = bulk_upsert_custom_entities.sync_detailed(client=self.client, json_body=body)
        return self._task_helper_from_response(response, BulkUpdateCustomEntitiesAsyncTaskResponse)



from typing import List, Optional

from benchling_api_client.v2.beta.api.folders import list_folder_collaborations
from benchling_api_client.v2.beta.models.collaboration import Collaboration
from benchling_api_client.v2.beta.models.collaborations_paginated_list import CollaborationsPaginatedList
from benchling_api_client.v2.beta.models.list_folder_collaborations_role import ListFolderCollaborationsRole
from benchling_api_client.v2.beta.models.list_folder_collaborations_sort import ListFolderCollaborationsSort
from benchling_api_client.v2.beta.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaFolderService(BaseService):
    """
    V2-Beta Folders.

    Folders are nested within projects to provide additional organization.

    See https://benchling.com/api/v2-beta/reference?showLa=true#/Folders
    """

    @api_method
    def _list_folder_collaborations_page(
        self,
        folder_id: str,
        user_id: Optional[str] = None,
        app_id: Optional[str] = None,
        team_id: Optional[str] = None,
        organization_id: Optional[str] = None,
        role: Optional[ListFolderCollaborationsRole] = None,
        ids: Optional[str] = None,
        modified_atlt: Optional[str] = None,
        modified_atgt: Optional[str] = None,
        modified_atlte: Optional[str] = None,
        modified_atgte: Optional[str] = None,
        created_atlt: Optional[str] = None,
        created_atgt: Optional[str] = None,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListFolderCollaborationsSort] = None,
    ) -> Response[CollaborationsPaginatedList]:
        response = list_folder_collaborations.sync_detailed(
            client=self.client,
            folder_id=folder_id,
            user_id=none_as_unset(user_id),
            app_id=none_as_unset(app_id),
            team_id=none_as_unset(team_id),
            organization_id=none_as_unset(organization_id),
            role=none_as_unset(role),
            ids=none_as_unset(optional_array_query_param(ids)),
            modified_atlt=none_as_unset(modified_atlt),
            modified_atlte=none_as_unset(modified_atlte),
            modified_atgt=none_as_unset(modified_atgt),
            modified_atgte=none_as_unset(modified_atgte),
            created_atlt=none_as_unset(created_atlt),
            created_atgt=none_as_unset(created_atgt),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            sort=none_as_unset(sort),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_collaborations(
        self,
        *,
        folder_id: str,
        user_id: Optional[str] = None,
        app_id: Optional[str] = None,
        team_id: Optional[str] = None,
        organization_id: Optional[str] = None,
        role: Optional[ListFolderCollaborationsRole] = None,
        ids: Optional[str] = None,
        modified_atlt: Optional[str] = None,
        modified_atgt: Optional[str] = None,
        modified_atlte: Optional[str] = None,
        modified_atgte: Optional[str] = None,
        created_atlt: Optional[str] = None,
        created_atgt: Optional[str] = None,
        page_size: Optional[int] = 50,
        sort: Optional[ListFolderCollaborationsSort] = None,
    ) -> PageIterator[Collaboration]:
        """
        List information about collaborations on the specified folder.

        See https://benchling.com/api/v2-beta/reference#/Folders/listFolderCollaborations
        """

        def api_call(next_token: NextToken) -> Response[CollaborationsPaginatedList]:
            return self._list_folder_collaborations_page(
                folder_id=folder_id,
                user_id=user_id,
                app_id=app_id,
                team_id=team_id,
                organization_id=organization_id,
                role=role,
                ids=ids,
                modified_atlt=modified_atlt,
                modified_atgt=modified_atgt,
                modified_atlte=modified_atlte,
                modified_atgte=modified_atgte,
                created_atlt=created_atlt,
                created_atgt=created_atgt,
                page_size=page_size,
                sort=_translate_to_string_enum(ListFolderCollaborationsSort, sort),
                next_token=next_token,
            )

        def results_extractor(body: CollaborationsPaginatedList) -> Optional[List[Collaboration]]:
            return body.collaborations

        return PageIterator(api_call, results_extractor)



from typing import Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.projects import (
    archive_projects,
    get_project,
    list_projects,
    unarchive_projects,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    ListProjectsSort,
    Project,
    ProjectsArchivalChange,
    ProjectsArchive,
    ProjectsArchiveReason,
    ProjectsPaginatedList,
    ProjectsUnarchive,
)
from benchling_sdk.services.v2.base_service import BaseService


class ProjectService(BaseService):
    """
    Projects.

    Manage project objects.

    See https://benchling.com/api/reference#/Projects
    """

    @api_method
    def get_by_id(self, project_id: str) -> Project:
        """
        Get a Project by ID.

        See https://benchling.com/api/reference#/Projects/getProject
        """
        response = get_project.sync_detailed(client=self.client, project_id=project_id)
        return model_from_detailed(response)

    @api_method
    def _projects_page(
        self,
        *,
        sort: Optional[ListProjectsSort] = ListProjectsSort.NAME,
        archive_reason: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[ProjectsPaginatedList]:
        response = list_projects.sync_detailed(
            client=self.client,
            sort=none_as_unset(sort),
            archive_reason=none_as_unset(archive_reason),
            ids=none_as_unset(optional_array_query_param(ids)),
            name=none_as_unset(name),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response

    def list(
        self,
        *,
        sort: Optional[Union[str, ListProjectsSort]] = None,
        archive_reason: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> PageIterator[Project]:
        """
        List Projects.

        See https://benchling.com/api/reference#/Projects/listProjects
        """

        def api_call(next_token: NextToken) -> Response[ProjectsPaginatedList]:
            return self._projects_page(
                sort=_translate_to_string_enum(ListProjectsSort, sort),
                archive_reason=archive_reason,
                ids=ids,
                name=name,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: ProjectsPaginatedList) -> Optional[List[Project]]:
            return body.projects

        return PageIterator(api_call, results_extractor)

    @api_method
    def archive(self, project_ids: Iterable[str], reason: ProjectsArchiveReason) -> ProjectsArchivalChange:
        """
        Archive Projects.

        See https://benchling.com/api/reference#/Projects/archiveProjects
        """
        archive_request = ProjectsArchive(project_ids=list(project_ids), reason=reason)
        response = archive_projects.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, project_ids: Iterable[str]) -> ProjectsArchivalChange:
        """
        Unarchive Projects.

        See https://benchling.com/api/reference#/Projects/unarchiveProjects
        """
        unarchive_request = ProjectsUnarchive(project_ids=list(project_ids))
        response = unarchive_projects.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)
